#ifndef __LIS_LEQLIB_H__
#define __LIS_LEQLIB_H__
/*************************************/
/* #define LIS_VERSION	"2005/07/01" */
/* #define LIS_VERSION	"2005/07/29" */
/* #define LIS_VERSION	"2005/09/08" */
#define LIS_VERSION	"2005/09/16"
/**********************************/
#include <stdio.h>
#include <stdarg.h>
#ifdef _MPI
	#include <mpi.h>
	typedef MPI_Comm			LIS_Comm;
	#define LIS_COMM_WORLD		((LIS_Comm)MPI_COMM_WORLD)
#else
	typedef int					LIS_Comm;
	#define LIS_COMM_WORLD		((LIS_Comm)0x1)
#endif


#define _max(a,b) ((a) >= (b) ? (a) : (b))
#define _min(a,b) ((a) <= (b) ? (a) : (b))

typedef double					LIS_SCALAR;

#define LIS_MATRIX_NULL		-12
#define LIS_MATRIX_ASSEMBLING	0
#define LIS_MATRIX_CRS			1
#define LIS_MATRIX_CSR			1
#define LIS_MATRIX_CCS			2
#define LIS_MATRIX_CSC			2
#define LIS_MATRIX_MSR			3
#define LIS_MATRIX_DIA			4
#define LIS_MATRIX_CDS			4
#define LIS_MATRIX_ELL			5
#define LIS_MATRIX_JDS			6
#define LIS_MATRIX_JAD			6
#define LIS_MATRIX_BSR			7
#define LIS_MATRIX_BSC			8
#define LIS_MATRIX_VBR			9
#define LIS_MATRIX_COO			10
#define LIS_MATRIX_DENSE		11
#define LIS_MATRIX_DNS			11

#define LIS_MATRIX_CRS_STR		"crs"
#define LIS_MATRIX_CCS_STR		"ccs"
#define LIS_MATRIX_MSR_STR		"msr"
#define LIS_MATRIX_DIA_STR		"dia"
#define LIS_MATRIX_ELL_STR		"ell"
#define LIS_MATRIX_JDS_STR		"jds"
#define LIS_MATRIX_BSR_STR		"bsr"
#define LIS_MATRIX_BSC_STR		"bsc"
#define LIS_MATRIX_VBR_STR		"vbr"
#define LIS_MATRIX_DNS_STR		"dns"
#define LIS_MATRIX_COO_STR		"coo"

#define LIS_PRECON_TYPE_LEN			9
#define LIS_PRECON_TYPE_NONE		0
#define LIS_PRECON_TYPE_JACOBI		1
#define LIS_PRECON_TYPE_ILU			2
#define LIS_PRECON_TYPE_SSOR		3
#define LIS_PRECON_TYPE_HYBRID		4
#define LIS_PRECON_TYPE_IS			5
#define LIS_PRECON_TYPE_SAI			6
#define LIS_PRECON_TYPE_SAAMG		7
#define LIS_PRECON_TYPE_RIC2S		8

#define LIS_MATRIX_ARRAY_BASE_0		0
#define LIS_MATRIX_ARRAY_BASE_1		1

#define LIS_FMT_AUTO				0
#define LIS_FMT_PLAIN				1
#define LIS_FMT_MM					2
#define LIS_FMT_LIS					3
#define LIS_FMT_LIS_ASCII			3
#define LIS_FMT_LIS_BINARY			4
#define LIS_FMT_FREE				5
#define LIS_FMT_ITBL				6
#define LIS_FMT_HB					7

#define LIS_BINARY_BIG				0
#define LIS_BINARY_LITTLE			1


#define LIS_PARAMS_LEN				11
#define LIS_PARAMS_RESID			0
#define LIS_PARAMS_W				1
#define LIS_PARAMS_RELAX			2
#define LIS_PARAMS_DROP				3
#define LIS_PARAMS_ALPHA			4
#define LIS_PARAMS_TAU				5
#define LIS_PARAMS_SIGMA			6
#define LIS_PARAMS_GAMMA			7
#define LIS_PARAMS_SSOR_W			8
#define LIS_PARAMS_PRESID			9
#define LIS_PARAMS_PW				10

#define LIS_OPTIONS_LEN				15
#define LIS_OPTIONS_SOLVER			0
#define LIS_OPTIONS_PRECON			1
#define LIS_OPTIONS_MAXITER			2
#define LIS_OPTIONS_OUTPUT			3
#define LIS_OPTIONS_RESTART			4
#define LIS_OPTIONS_ELL				5
#define LIS_OPTIONS_SCALE			6
#define LIS_OPTIONS_FILL			7
#define LIS_OPTIONS_M				8
#define LIS_OPTIONS_PSOLVER			9
#define LIS_OPTIONS_PMAXITER		10
#define LIS_OPTIONS_PRESTART		11
#define LIS_OPTIONS_PELL			12
#define LIS_OPTIONS_PADDING			13
#define LIS_OPTIONS_ISLEVEL			14

#define LIS_STATUS_LEN				6
#define LIS_STATUS_RET				0
#define LIS_STATUS_ITER				1
#define LIS_STATUS_RESID			2
#define LIS_STATUS_TIMES			3
#define LIS_STATUS_PTIMES			4
#define LIS_STATUS_ITIMES			5

#define LIS_PRINT_NONE				0
#define LIS_PRINT_MEM				1
#define LIS_PRINT_OUT				2
#define LIS_PRINT_ALL				3

#define LIS_SCALE_NONE				0
#define LIS_SCALE_JACOBI			1
#define LIS_SCALE_SYMM_DIAG			2

#define LIS_SOLVER_LEN				12
#define LIS_SOLVER_CG				1
#define LIS_SOLVER_BICG				2
#define LIS_SOLVER_CGS				3
#define LIS_SOLVER_BICGSTAB			4
#define LIS_SOLVER_BICGSTABL		5
#define LIS_SOLVER_GPBICG			6
#define LIS_SOLVER_ORTHOMIN			7
#define LIS_SOLVER_GMRES			8
#define LIS_SOLVER_QMR				9
#define LIS_SOLVER_JACOBI			10
#define LIS_SOLVER_GS				11
#define LIS_SOLVER_SOR				12


typedef struct
{
	LIS_Comm	comm;
	int			neibpetot;
	int			imnnz;
	int			exnnz;
	int			*neibpe;
	int			*index_export;
	int			*node_export;
	int			*index_import;
	int			*node_import;
	LIS_SCALAR	*ws;
	LIS_SCALAR	*wr;
#ifdef _MPI
	MPI_Request *req1,*req2;
	MPI_Status	*sta1,*sta2;
#endif
} LIS_COMMTABLE;

/*        */
/* Matrix */
/*        */
struct LIS_MATRIX_STRUCT
{
	int			matrix_type;
	int			gn;
	int			n;
	int			np;
	int			nnz;
	int			ndz;
	int			bn;
	int			bnr;
	int			bnc;
	int			nr;
	int			nc;
	int			bnnz;
	int			nnd;
	int			maxnzr;
	int			*ptr;
	int			*row;
	int			*col;
	int			*index;
	int			*bptr;
	int			*bindex;
	LIS_SCALAR	*value;
	LIS_SCALAR	*work;

	int			my_rank;
	int			nprocs;
	LIS_Comm	comm;
	int			is;
	int			ie;
	int			*g_index;
	int			*g_col;
	int			*ranges;
	LIS_COMMTABLE	*commtable;
};
typedef struct LIS_MATRIX_STRUCT *LIS_MATRIX;

typedef struct
{
	int			matrix_type;
	int			gn;
	int			np;
	int			n;
	int			bn;
	int			nnz;
	int			ndz;
	int			bnr;
	int			bnc;
	int			nr;
	int			nc;
	int			bnnz;
	int			nnd;
	int			maxnzr;
} LIS_MATRIX_PARAMS;

/*                */
/* Vector         */
/*                */
struct LIS_VECTOR_STRUCT
{
	int			gn;
	int			n;
	int			np;
	LIS_SCALAR	*value;

	int			my_rank;
	int			nprocs;
	LIS_Comm	comm;
	int			is;
	int			ie;
	int			*g_index;
	int			*ranges;
};
typedef struct LIS_VECTOR_STRUCT *LIS_VECTOR;

/*                */
/* Preconditioner */
/*                */
struct LIS_PRECON_STRUCT
{
	int			precon_type;
	LIS_MATRIX	Pmat;
	LIS_MATRIX	Amat;
	LIS_MATRIX	At;
	LIS_MATRIX	L;
	LIS_MATRIX	U;
	LIS_VECTOR	*work;
	LIS_VECTOR	D;
	LIS_VECTOR	Pb;
	LIS_VECTOR	temp;
	LIS_SCALAR	*WS,*WR;
	LIS_SCALAR	*params;
	LIS_SCALAR	status[LIS_STATUS_LEN];
	int			*options;
	int			*diag;
	int			solver_comm;
	int			level_num;
	int			wsize;
	int			my_rank;
	int			nprocs;
	int			is_copy;
	int			worklen;
	struct LIS_PRECON_STRUCT *precon;
};
typedef struct LIS_PRECON_STRUCT *LIS_PRECON;

#ifdef __cplusplus
extern "C"
{
#endif

/****************************/
/* Vector                   */
/****************************/
	extern int lis_vector_create(int local_n, int global_n, LIS_Comm comm, LIS_VECTOR *vec);
	extern int lis_vector_destroy(LIS_VECTOR vec);
	extern int lis_vector_duplicate(LIS_VECTOR vin, LIS_VECTOR *vout);
	extern int lis_vector_duplicateM(LIS_MATRIX A, LIS_VECTOR *vout);
	extern int lis_vector_set(int i, LIS_SCALAR value, LIS_VECTOR v);
	extern int lis_vector_set_all(LIS_SCALAR alpha, LIS_VECTOR vx);
	extern int lis_vector_get_range(LIS_VECTOR v, int *is, int *ie);
	extern int lis_vector_copy(LIS_VECTOR vsrc, LIS_VECTOR vdst);
	extern int lis_vector_scale(LIS_SCALAR alpha, LIS_VECTOR vx);
	extern int lis_vector_dot(LIS_VECTOR vx, LIS_VECTOR vy, LIS_SCALAR *val);
	extern int lis_vector_nrm2(LIS_VECTOR vx, LIS_SCALAR *val);
	extern int lis_vector_axpy(LIS_SCALAR alpha, LIS_VECTOR vx, LIS_VECTOR vy);
	extern int lis_vector_xpay(LIS_VECTOR vx, LIS_SCALAR alpha, LIS_VECTOR vy);
	extern int lis_vector_axpyz(LIS_SCALAR alpha, LIS_VECTOR vx, LIS_VECTOR vy, LIS_VECTOR vz);
	extern void lis_vector_print(LIS_VECTOR x);
/****************************/
/* Matrix                   */
/****************************/
	extern int lis_matrix_create(int local_n, int global_n, LIS_Comm comm, LIS_MATRIX *Amat);
	extern int lis_matrix_get_range(LIS_MATRIX A, int *is, int *ie);
	extern int lis_matrix_set(int i, int j, LIS_SCALAR value, LIS_MATRIX A);
	extern int lis_matrix_set_crs(int nnz, int *row, int *index, LIS_SCALAR *value, LIS_MATRIX A);
	extern int lis_matrix_set_ccs(int nnz, int *col, int *index, LIS_SCALAR *value, LIS_MATRIX A);
	extern int lis_matrix_set_msr(int nnz, int ndz, int *index, LIS_SCALAR *value, LIS_MATRIX A);
#ifdef _OPENMP
	extern int lis_matrix_set_dia(int *nnd, int *index, LIS_SCALAR *value, LIS_MATRIX A);
#else
	extern int lis_matrix_set_dia(int nnd, int *index, LIS_SCALAR *value, LIS_MATRIX A);
#endif
	extern int lis_matrix_set_ell(int maxnzr, int *index, LIS_SCALAR *value, LIS_MATRIX A);
#ifdef _OPENMP
	extern int lis_matrix_set_jds(int nnz, int *maxnzr, int *perm, int *ptr, int *index, LIS_SCALAR *value, LIS_MATRIX A);
#else
	extern int lis_matrix_set_jds(int nnz, int maxnzr, int *perm, int *ptr, int *index, LIS_SCALAR *value, LIS_MATRIX A);
#endif
	extern int lis_matrix_set_bsr(int bnr, int bnc, int bnnz, int *bptr, int *bindex, LIS_SCALAR *value, LIS_MATRIX A);
	extern int lis_matrix_set_bsc(int bnr, int bnc, int bnnz, int *bptr, int *bindex, LIS_SCALAR *value, LIS_MATRIX A);
	extern int lis_matrix_set_vbr(int nnz, int nr, int nc, int bnnz, int *row, int *col, int *index, int *bptr, int *bindex, LIS_SCALAR *value, LIS_MATRIX A);
	extern int lis_matrix_set_coo(int nnz, int *row, int *col, LIS_SCALAR *value, LIS_MATRIX A);
	extern int lis_matrix_set_dns(LIS_SCALAR *value, LIS_MATRIX A);
	extern int lis_matrix_assemble(LIS_MATRIX *A, int matrix_type, ...);
	extern int lis_matrix_duplicate(LIS_MATRIX Ain, LIS_MATRIX *Aout);
	extern int lis_matrix_destroy(LIS_MATRIX Amat);
	extern int lis_matrix_print(LIS_MATRIX Amat);
	extern int lis_matrix_split(LIS_MATRIX A, LIS_VECTOR *D, LIS_MATRIX *L, LIS_MATRIX *U);
	extern int lis_matrix_get_diagonal(LIS_MATRIX A, LIS_VECTOR d);
	extern int lis_matrix_scaling(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR d, int action);
	extern int lis_matrix_convert(LIS_MATRIX Ain, LIS_MATRIX *Aout, int convert_matrix_type, ...);
	extern int lis_matrix_sets(int i[], int j[], int value[], LIS_MATRIX A);
	extern int lis_matrix_copy(LIS_MATRIX Ain, LIS_MATRIX *Aout);
/****************************/
/* Matrix Vector Product    */
/****************************/
	extern void lis_matvec(LIS_MATRIX A, LIS_VECTOR x, LIS_VECTOR y);
	extern void lis_matvect(LIS_MATRIX A, LIS_VECTOR x, LIS_VECTOR y);
/****************************/
/* Solvers                  */
/****************************/
	extern int lis_solve(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x, LIS_SCALAR params[], int options[], LIS_SCALAR status[]);
	extern void lis_get_solve_status(LIS_SCALAR status[], int *returnval, int *iter, LIS_SCALAR *residual, LIS_SCALAR *times);
/****************************/
/* I/O                      */
/****************************/
	extern int lis_input(LIS_MATRIX *A, LIS_VECTOR *b, LIS_VECTOR *x, char *filename, LIS_Comm comm, int matrix_type, ...);
	extern int lis_output(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x, int mode, char *path);
	extern int lis_output_vector(LIS_VECTOR v, int format, char *filename);
	extern int lis_input_vector(LIS_VECTOR *v, char *filename, LIS_Comm comm);
	extern int lis_output_residual_history(char *filename);
/****************************/
/* Utilities                */
/****************************/
	extern int lis_initialize(int argc, char* argv[], char *fname, int options[], LIS_SCALAR params[]);
	extern int lis_finalize();
	extern int lis_set_options(char *text, int options[], LIS_SCALAR params[]);
	extern int lis_get_residual_history(LIS_VECTOR v);
	extern int lis_get_solvername(int sol, char *name);
	extern void *lis_malloc( size_t size);
	extern void *lis_realloc( void *p, size_t size );
	extern void lis_free(void *p);
	extern void lis_free2(int n, ...);
	extern double lis_wtime(void);
	extern void lis_sort_i(int is, int ie, int *i1);
	extern void lis_sort_ii(int is, int ie, int *i1, int *i2);
	extern void lis_sort_id(int is, int ie, int *i1, LIS_SCALAR *d1);
	extern void lis_sort_iid(int is, int ie, int *i1, int *i2, LIS_SCALAR *d1);
	extern void lis_sort_iiid(int is, int ie, int *i1, int *i2, int *i3, LIS_SCALAR *d1);
	extern void lis_sortr_ii(int is, int ie, int *i1, int *i2);

	extern void CHKERR(int err);
	extern void SETERR(char *fmt, ...);


#ifdef __cplusplus
}
#endif

#define LIS_TRUE							1
#define LIS_FALSE							0
#define LIS_FAILS							-1
#define LIS_SUCCESS							0
#define LIS_ILL_OPTION						1
#define LIS_BREAKDOWN						2
#define LIS_OUT_OF_MEMORY					3
#define LIS_MAXITER							4
#define LIS_DIMENSION_LE_ZERO				5
#define LIS_SOLVERS_UNKNOWN					5
#define LIS_MAXITER_LE_ZERO					6
#define LIS_CONVERGENCE						7
#define LIS_NOT_IMPLEMENTED					13 


#define LIS_GET_ISIE(id,nprocs,n,is,ie) \
		if( id < n%nprocs )				\
		{								\
			ie = n/nprocs+1;			\
			is = ie*id;					\
		}								\
		else							\
		{								\
			ie = n/nprocs;				\
			is = ie*id + n%nprocs;		\
		}								\
		ie = ie+is;

#endif
