#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <ctype.h>
#include <string.h>
#include <stdarg.h>
#ifdef _OPENMP
	#include <omp.h>
#endif
#include "lis.h"
#include "lis_io.h"

int lis_input_lis(LIS_MATRIX *A, LIS_VECTOR *b, LIS_VECTOR *x, char *filename, FILE *file, LIS_Comm comm, int matrix_type, ...)
{
	va_list			vvlist;
	int				bnr,bnc;
	int				*crow,*ccol;
	char			buf[BUFSIZE],banner[128],mode[128],mattype[128];
	int				in_matrix_type, in_mode;
	int				err;
	LIS_MATRIX		Ain;

	va_start( vvlist, matrix_type );
	bnr = va_arg(vvlist, int);
	bnc = va_arg(vvlist, int);
	crow = va_arg(vvlist, int *);
	ccol = va_arg(vvlist, int *);
	va_end(vvlist);

	if( fgets(buf, BUFSIZE, file) == NULL )
	{
		SETERR("file i/o error\n");
		return LIS_FAILS;
	}
	buf[10] = '\0';
	sscanf(buf, "%s %s %s", banner, mode, mattype);
	if( strncmp(banner, LISBanner, strlen(LISBanner)) != 0)
	{
		return LIS_FAILS;
	}

	in_mode = LIS_FMT_LIS_ASCII;
	if( mode[0]=='B' || mode[0]=='L' )
	{
		fclose(file);
		file = fopen(filename, "rb");
		if( file==NULL )
		{
			printf("cannot open file %s\n", filename);
			return LIS_FAILS;
		}
		fread(buf, sizeof(char), 10, file);
		in_mode = 1;
		in_mode = *(char *)&in_mode;
		if( (in_mode==LIS_BINARY_BIG && mode[0]=='L') || (in_mode==LIS_BINARY_LITTLE && mode[0]=='B') )
		{
			in_mode = LIS_TRUE;
		}
		else
		{
			in_mode = LIS_FALSE;
		}
	}

	if( strncmp(mattype, LIS_MATRIX_CRS_STR, strlen(LIS_MATRIX_CRS_STR)) == 0 )
	{
		in_matrix_type = LIS_MATRIX_CRS;
	}
	else if( strncmp(mattype, LIS_MATRIX_CCS_STR, strlen(LIS_MATRIX_CCS_STR)) == 0 )
	{
		in_matrix_type = LIS_MATRIX_CCS;
	}
	else if( strncmp(mattype, LIS_MATRIX_MSR_STR, strlen(LIS_MATRIX_MSR_STR)) == 0 )
	{
		in_matrix_type = LIS_MATRIX_MSR;
	}
	else if( strncmp(mattype, LIS_MATRIX_DIA_STR, strlen(LIS_MATRIX_DIA_STR)) == 0 )
	{
		in_matrix_type = LIS_MATRIX_DIA;
	}
	else if( strncmp(mattype, LIS_MATRIX_ELL_STR, strlen(LIS_MATRIX_ELL_STR)) == 0 )
	{
		in_matrix_type = LIS_MATRIX_ELL;
	}
	else if( strncmp(mattype, LIS_MATRIX_JDS_STR, strlen(LIS_MATRIX_JDS_STR)) == 0 )
	{
		in_matrix_type = LIS_MATRIX_JDS;
	}
	else if( strncmp(mattype, LIS_MATRIX_BSR_STR, strlen(LIS_MATRIX_BSR_STR)) == 0 )
	{
		in_matrix_type = LIS_MATRIX_BSR;
	}
	else if( strncmp(mattype, LIS_MATRIX_BSC_STR, strlen(LIS_MATRIX_BSC_STR)) == 0 )
	{
		in_matrix_type = LIS_MATRIX_BSC;
	}
	else if( strncmp(mattype, LIS_MATRIX_VBR_STR, strlen(LIS_MATRIX_VBR_STR)) == 0 )
	{
		in_matrix_type = LIS_MATRIX_VBR;
	}
	else if( strncmp(mattype, LIS_MATRIX_DNS_STR, strlen(LIS_MATRIX_DNS_STR)) == 0 )
	{
		in_matrix_type = LIS_MATRIX_DNS;
	}
	else if( strncmp(mattype, LIS_MATRIX_COO_STR, strlen(LIS_MATRIX_COO_STR)) == 0 )
	{
		in_matrix_type = LIS_MATRIX_COO;
	}


#ifdef _MPI
	MPI_Barrier(comm);
#endif
	switch( in_matrix_type )
	{
	case LIS_MATRIX_CRS:
		err = lis_input_lis_crs(&Ain,b,x,in_mode,file,comm);
		break;
	case LIS_MATRIX_CCS:
		err = lis_input_lis_ccs(&Ain,b,x,in_mode,file,comm);
		break;
	case LIS_MATRIX_MSR:
		err = lis_input_lis_msr(&Ain,b,x,in_mode,file,comm);
		break;
	case LIS_MATRIX_DIA:
		err = lis_input_lis_dia(&Ain,b,x,in_mode,file,comm);
		break;
	case LIS_MATRIX_ELL:
		err = lis_input_lis_ell(&Ain,b,x,in_mode,file,comm);
		break;
	case LIS_MATRIX_JDS:
		err = lis_input_lis_jds(&Ain,b,x,in_mode,file,comm);
		break;
	case LIS_MATRIX_BSR:
		err = lis_input_lis_bsr(&Ain,b,x,in_mode,file,comm);
		break;
	case LIS_MATRIX_BSC:
		err = lis_input_lis_bsc(&Ain,b,x,in_mode,file,comm);
		break;
	case LIS_MATRIX_VBR:
		err = lis_input_lis_vbr(&Ain,b,x,in_mode,file,comm);
		break;
	case LIS_MATRIX_COO:
		err = lis_input_lis_coo(&Ain,b,x,in_mode,file,comm);
		break;
	case LIS_MATRIX_DNS:
		err = lis_input_lis_dns(&Ain,b,x,in_mode,file,comm);
		break;
	}
	if( in_matrix_type!=matrix_type )
	{
		#ifdef _MPI
		MPI_Barrier(comm);
		#endif
		lis_matrix_convert(Ain,A,matrix_type,bnr,bnc,crow,ccol);
		lis_matrix_destroy(Ain);
	}
	else
	{
		*A = Ain;
	}
	return LIS_SUCCESS;
}

static char seps[] = " ,\t\n\r";
int lis_fscan_int(int n, FILE *file, int val[], int origin)
{
	char p[256];
	char *token;
	int i;

	i=0;
	while( i<n )
	{
		if( fgets(p, 256, file)==NULL )
		{
			return LIS_FAILS;
		}
		token = strtok( p, seps );
		while( token!=NULL )
		{
			sscanf(token, "%d", &val[i]);
			val[i++] -= origin;
			token = strtok( NULL, seps );
		}
	}
	return LIS_SUCCESS;
}

int lis_fscan_double(int n, FILE *file, double val[])
{
/*	char p[256];
	char *token;*/
	int i,k;

	i=0;
	while( i<n )
	{
		fscanf(file, "%lg", &val[i++]);
		/*
		if( fgets(p, 256, file)==NULL )
		{
			return LIS_FAILS;
		}
		token = strtok( p, seps );
		while( token!=NULL )
		{
			k = sscanf(token, "%lg", &val[i++]);
			while( k!=1 )
			{
				token += k;
				k = sscanf(token, "%lg", &val[i++]);
			}
			token = strtok( NULL, seps );
		}
		*/
	}
	return LIS_SUCCESS;
}

int lis_bswap_int(int n, int *buf)
{
	int i;
	int t;
	int *p;
	char *pp, *pt;

	p = buf;
	for(i=0;i<n;i++)
	{
		t = *p;
		pp = (char *)p;
		pt = (char *)&t;
		pp[0] = pt[3];
		pp[1] = pt[2];
		pp[2] = pt[1];
		pp[3] = pt[0];
		p++;
	}
	return LIS_SUCCESS;
}

int lis_bswap_double(int n, double *buf)
{
	int i;
	double t;
	double *p;
	char *pp, *pt;

	p = buf;
	for(i=0;i<n;i++)
	{
		t = *p;
		pp = (char *)p;
		pt = (char *)&t;
		pp[0] = pt[7];
		pp[1] = pt[6];
		pp[2] = pt[5];
		pp[3] = pt[4];
		pp[4] = pt[3];
		pp[5] = pt[2];
		pp[6] = pt[1];
		pp[7] = pt[0];
		p++;
	}
	return LIS_SUCCESS;
}

int lis_bswap_size_t(int n, size_t *buf)
{
	int i;
	size_t t;
	size_t *p;
	char *pp, *pt;

	p = buf;
	for(i=0;i<n;i++)
	{
		t = *p;
		pp = (char *)p;
		pt = (char *)&t;
		pp[0] = pt[7];
		pp[1] = pt[6];
		pp[2] = pt[5];
		pp[3] = pt[4];
		pp[4] = pt[3];
		pp[5] = pt[2];
		pp[6] = pt[1];
		pp[7] = pt[0];
		p++;
	}
	return LIS_SUCCESS;
}

int lis_input_lis_crs(LIS_MATRIX *A, LIS_VECTOR *b, LIS_VECTOR *x, int mode, FILE *file, LIS_Comm comm)
{
	int				n,nnz,nprocs,my_rank,pe;
	int				i;
	int				err;
	int				ibuf[10];
	int				*ptr,*index;
	char			cbuf[BUFSIZE];
	char			c;
	char			*p;
	size_t			*tbuf;
	LIS_SCALAR		*value;

#ifdef _MPI
	MPI_Comm_size(comm,&nprocs);
	MPI_Comm_rank(comm,&my_rank);
#else
	nprocs  = 1;
	my_rank = 0;
#endif
	if( mode==LIS_FMT_LIS_ASCII )
	{
		if( fgets(cbuf, BUFSIZE, file) == NULL )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( sscanf(cbuf, "%d %d %d",&ibuf[0], &ibuf[1], &ibuf[2])!=3 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( nprocs!=ibuf[2] )
		{
			SETERR("The number of PE=(%d) is different (in file PE=%d)\n",nprocs,ibuf[2]);
			return LIS_FAILS;
		}
		pe=-1;
		do
		{
			if( fgets(cbuf, BUFSIZE, file) == NULL )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
			if( cbuf[0]=='#' )
			{
				if( sscanf(cbuf, "%c %d %d %d",&c, &pe, &ibuf[2], &ibuf[3])!=4 )
				{
					SETERR("file i/o error\n");
					return LIS_FAILS;
				}
			}
		}while(pe!=my_rank);
	}
	else
	{
		if( fread(ibuf, sizeof(int), 3, file)!=3 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( mode )
		{
			lis_bswap_int(3,(int *)ibuf);
		}
		if( nprocs!=ibuf[2] )
		{
			SETERR("The number of PE=(%d) is different (in file PE=%d)\n",nprocs,ibuf[2]);
			return LIS_FAILS;
		}
		tbuf = (size_t *)lis_malloc( (nprocs+1)*sizeof(size_t) );
		if( tbuf==NULL )
		{
			SETERR("out of memory\n");
			return LIS_OUT_OF_MEMORY;
		}
		if( fread(tbuf, sizeof(size_t), nprocs+1, file)!=nprocs+1 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( mode )
		{
			lis_bswap_size_t(nprocs+1,(size_t *)tbuf);
		}
		fseek(file,tbuf[my_rank],SEEK_SET);

		if( fread(&ibuf[2], sizeof(int), 2, file)!=2 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( mode )
		{
			lis_bswap_int(2,(int *)ibuf);
		}
		lis_free(tbuf);
	}

	n   = ibuf[2];
	nnz = ibuf[3];
	err = lis_matrix_create(n,0,comm,A);
	if( err )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	if( ibuf[0] )
	{
		err = lis_vector_duplicateM(*A,b);
		if( err )
		{
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
	}
	if( ibuf[1] )
	{
		err = lis_vector_duplicateM(*A,x);
		if( err )
		{
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
	}

	if( mode==LIS_FMT_LIS_ASCII )
	{
		// ptr
		err = lis_fscan_int(n+1,file,ptr,0);
		if( err )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// index
		err = lis_fscan_int(nnz,file,index,0);
		if( err )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// value
		err = lis_fscan_double(nnz,file,value);
		if( err )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// RHS
		if( ibuf[0] )
		{
			err = lis_fscan_double(n,file,(*b)->value);
			if( err )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
		}
		// solution
		if( ibuf[1] )
		{
			err = lis_fscan_double(n,file,(*x)->value);
			if( err )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
		}
	}
	else
	{
		// ptr
		if( fread(ptr, sizeof(int), n+1, file)!=n+1 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// index
		if( fread(index, sizeof(int), nnz, file)!=nnz )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// value
		if( fread(value, sizeof(LIS_SCALAR), nnz, file)!=nnz )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// RHS
		if( ibuf[0] )
		{
			if( fread((*b)->value, sizeof(LIS_SCALAR), n, file)!=n )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
		}
		// solution
		if( ibuf[1] )
		{
			if( fread((*x)->value, sizeof(LIS_SCALAR), n, file)!=n )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
		}

		if( mode )
		{
			lis_bswap_int(n+1,ptr);
			lis_bswap_int(nnz,index);
			lis_bswap_double(nnz,value);
			if( ibuf[0] ) lis_bswap_double(n,(*b)->value);
			if( ibuf[1] ) lis_bswap_double(n,(*x)->value);
		}
	}
	err = lis_matrix_set_crs(nnz,ptr,index,value,*A);
	if( err ) return err;
	err = lis_matrix_assemble(A,LIS_MATRIX_CRS);
	if( err ) return err;

	return LIS_SUCCESS;
}

int lis_input_lis_ccs(LIS_MATRIX *A, LIS_VECTOR *b, LIS_VECTOR *x, int mode, FILE *file, LIS_Comm comm)
{
	int				n,nnz,nprocs,my_rank,pe;
	int				i;
	int				err;
	int				ibuf[10];
	int				*ptr,*index;
	char			cbuf[BUFSIZE];
	char			c;
	char			*p;
	size_t			*tbuf;
	LIS_SCALAR		*value;

#ifdef _MPI
	MPI_Comm_size(comm,&nprocs);
	MPI_Comm_rank(comm,&my_rank);
#else
	nprocs  = 1;
	my_rank = 0;
#endif
	if( mode==LIS_FMT_LIS_ASCII )
	{
		if( fgets(cbuf, BUFSIZE, file) == NULL )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( sscanf(cbuf, "%d %d %d",&ibuf[0], &ibuf[1], &ibuf[2])!=3 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( nprocs!=ibuf[2] )
		{
			SETERR("The number of PE=(%d) is different (in file PE=%d)\n",nprocs,ibuf[2]);
			return LIS_FAILS;
		}
		pe=-1;
		do
		{
			if( fgets(cbuf, BUFSIZE, file) == NULL )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
			if( cbuf[0]=='#' )
			{
				if( sscanf(cbuf, "%c %d %d %d",&c, &pe, &ibuf[2], &ibuf[3])!=4 )
				{
					SETERR("file i/o error\n");
					return LIS_FAILS;
				}
			}
		}while(pe!=my_rank);
	}
	else
	{
		if( fread(ibuf, sizeof(int), 3, file)!=3 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( mode )
		{
			lis_bswap_int(3,(int *)ibuf);
		}
		if( nprocs!=ibuf[2] )
		{
			SETERR("The number of PE=(%d) is different (in file PE=%d)\n",nprocs,ibuf[2]);
			return LIS_FAILS;
		}
		tbuf = (size_t *)lis_malloc( (nprocs+1)*sizeof(size_t) );
		if( tbuf==NULL )
		{
			SETERR("out of memory\n");
			return LIS_OUT_OF_MEMORY;
		}
		if( fread(tbuf, sizeof(size_t), nprocs+1, file)!=nprocs+1 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( mode )
		{
			lis_bswap_size_t(nprocs+1,(size_t *)tbuf);
		}
		fseek(file,tbuf[my_rank],SEEK_SET);

		if( fread(&ibuf[2], sizeof(int), 2, file)!=2 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( mode )
		{
			lis_bswap_int(2,(int *)ibuf);
		}
		lis_free(tbuf);
	}

	n   = ibuf[2];
	nnz = ibuf[3];
	err = lis_matrix_create(n,0,comm,A);
	if( err )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	if( ibuf[0] )
	{
		err = lis_vector_duplicateM(*A,b);
		if( err )
		{
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
	}
	if( ibuf[1] )
	{
		err = lis_vector_duplicateM(*A,x);
		if( err )
		{
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
	}

	if( mode==LIS_FMT_LIS_ASCII )
	{
		// ptr
		err = lis_fscan_int(n+1,file,ptr,0);
		if( err )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// index
		err = lis_fscan_int(nnz,file,index,0);
		if( err )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// value
		err = lis_fscan_double(nnz,file,value);
		if( err )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// RHS
		if( ibuf[0] )
		{
			err = lis_fscan_double(n,file,(*b)->value);
			if( err )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
		}
		// solution
		if( ibuf[1] )
		{
			err = lis_fscan_double(n,file,(*x)->value);
			if( err )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
		}
	}
	else
	{
		// ptr
		if( fread(ptr, sizeof(int), n+1, file)!=n+1 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// index
		if( fread(index, sizeof(int), nnz, file)!=nnz )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// value
		if( fread(value, sizeof(LIS_SCALAR), nnz, file)!=nnz )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// RHS
		if( ibuf[0] )
		{
			if( fread((*b)->value, sizeof(LIS_SCALAR), n, file)!=n )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
		}
		// solution
		if( ibuf[1] )
		{
			if( fread((*x)->value, sizeof(LIS_SCALAR), n, file)!=n )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
		}

		if( mode )
		{
			lis_bswap_int(n+1,ptr);
			lis_bswap_int(nnz,index);
			lis_bswap_double(nnz,value);
			if( ibuf[0] ) lis_bswap_double(n,(*b)->value);
			if( ibuf[1] ) lis_bswap_double(n,(*x)->value);
		}
	}
	err = lis_matrix_set_ccs(nnz,ptr,index,value,*A);
	if( err ) return err;
	err = lis_matrix_assemble(A,LIS_MATRIX_CCS);
	if( err ) return err;

	return LIS_SUCCESS;
}

int lis_input_lis_msr(LIS_MATRIX *A, LIS_VECTOR *b, LIS_VECTOR *x, int mode, FILE *file, LIS_Comm comm)
{
	int				n,nnz,ndz,nprocs,my_rank,pe;
	int				i;
	int				err;
	int				ibuf[10];
	int				*ptr,*index;
	char			cbuf[BUFSIZE];
	char			c;
	char			*p;
	size_t			*tbuf;
	LIS_SCALAR		*value;

#ifdef _MPI
	MPI_Comm_size(comm,&nprocs);
	MPI_Comm_rank(comm,&my_rank);
#else
	nprocs  = 1;
	my_rank = 0;
#endif
	if( mode==LIS_FMT_LIS_ASCII )
	{
		if( fgets(cbuf, BUFSIZE, file) == NULL )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( sscanf(cbuf, "%d %d %d",&ibuf[0], &ibuf[1], &ibuf[2])!=3 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( nprocs!=ibuf[2] )
		{
			SETERR("The number of PE=(%d) is different (in file PE=%d)\n",nprocs,ibuf[2]);
			return LIS_FAILS;
		}

		pe=-1;
		do
		{
			if( fgets(cbuf, BUFSIZE, file) == NULL )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
			if( cbuf[0]=='#' )
			{
				if( sscanf(cbuf, "%c %d %d %d %d",&c, &pe, &ibuf[2], &ibuf[3], &ibuf[4])!=5 )
				{
					SETERR("file i/o error\n");
					return LIS_FAILS;
				}
			}
		}while(pe!=my_rank);
	}
	else
	{
		if( fread(ibuf, sizeof(int), 3, file)!=3 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( mode )
		{
			lis_bswap_int(3,(int *)ibuf);
		}
		if( nprocs!=ibuf[2] )
		{
			SETERR("The number of PE=(%d) is different (in file PE=%d)\n",nprocs,ibuf[2]);
			return LIS_FAILS;
		}
		tbuf = (size_t *)lis_malloc( (nprocs+1)*sizeof(size_t) );
		if( tbuf==NULL )
		{
			SETERR("out of memory\n");
			return LIS_OUT_OF_MEMORY;
		}
		if( fread(tbuf, sizeof(size_t), nprocs+1, file)!=nprocs+1 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( mode )
		{
			lis_bswap_size_t(nprocs+1,(size_t *)tbuf);
		}
		fseek(file,tbuf[my_rank],SEEK_SET);

		if( fread(&ibuf[2], sizeof(int), 3, file)!=3 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( mode )
		{
			lis_bswap_int(3,(int *)ibuf);
		}
		lis_free(tbuf);
	}

	n   = ibuf[2];
	nnz = ibuf[3];
	ndz = ibuf[4];
	err = lis_matrix_create(n,0,comm,A);
	if( err )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( (nnz+ndz+1)*sizeof(int) );
	if( index==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( (nnz+ndz+1)*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	if( ibuf[0] )
	{
		err = lis_vector_duplicateM(*A,b);
		if( err )
		{
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
	}
	if( ibuf[1] )
	{
		err = lis_vector_duplicateM(*A,x);
		if( err )
		{
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
	}

	if( mode==LIS_FMT_LIS_ASCII )
	{
		// index
		err = lis_fscan_int(nnz+ndz+1,file,index,0);
		if( err )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// value
		err = lis_fscan_double(nnz+ndz+1,file,value);
		if( err )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// RHS
		if( ibuf[0] )
		{
			err = lis_fscan_double(n,file,(*b)->value);
			if( err )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
		}
		// solution
		if( ibuf[1] )
		{
			err = lis_fscan_double(n,file,(*x)->value);
			if( err )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
		}
	}
	else
	{
		// index
		if( fread(index, sizeof(int), nnz+ndz+1, file)!=nnz+ndz+1 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// value
		if( fread(value, sizeof(LIS_SCALAR), nnz+ndz+1, file)!=nnz+ndz+1 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// RHS
		if( ibuf[0] )
		{
			if( fread((*b)->value, sizeof(LIS_SCALAR), n, file)!=n )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
		}
		// solution
		if( ibuf[1] )
		{
			if( fread((*x)->value, sizeof(LIS_SCALAR), n, file)!=n )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
		}

		if( mode )
		{
			lis_bswap_int(nnz+ndz+1,index);
			lis_bswap_double(nnz+ndz+1,value);
			if( ibuf[0] ) lis_bswap_double(n,(*b)->value);
			if( ibuf[1] ) lis_bswap_double(n,(*x)->value);
		}
	}
	err = lis_matrix_set_msr(nnz,ndz,index,value,*A);
	if( err ) return err;
	err = lis_matrix_assemble(A,LIS_MATRIX_MSR);
	if( err ) return err;

	return LIS_SUCCESS;
}

int lis_input_lis_dia(LIS_MATRIX *A, LIS_VECTOR *b, LIS_VECTOR *x, int mode, FILE *file, LIS_Comm comm)
{
#ifdef _OPENMP
	int				n,nnd,nprocs,my_rank,pe;
	int				i;
	int				err;
	int				ibuf[10];
	int				*index,*col;
	char			cbuf[BUFSIZE];
	char			c;
	char			*p;
	size_t			*tbuf;
	LIS_SCALAR		*value;

	nprocs  = omp_get_max_threads();
	my_rank = 0;
	if( mode==LIS_FMT_LIS_ASCII )
	{
		if( fgets(cbuf, BUFSIZE, file) == NULL )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( sscanf(cbuf, "%d %d %d",&ibuf[0], &ibuf[1], &ibuf[2])!=3 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( nprocs!=ibuf[2] )
		{
			SETERR("The number of PE=(%d) is different (in file PE=%d)\n",nprocs,ibuf[2]);
			return LIS_FAILS;
		}

		pe=-1;
		do
		{
			if( fgets(cbuf, BUFSIZE, file) == NULL )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
			if( cbuf[0]=='#' )
			{
				if( sscanf(cbuf, "%c %d %d %d",&c, &pe, &ibuf[2], &ibuf[3])!=4 )
				{
					SETERR("file i/o error\n");
					return LIS_FAILS;
				}
			}
		}while(pe!=my_rank);
	}
	else
	{
		if( fread(ibuf, sizeof(int), 3, file)!=3 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( mode )
		{
			lis_bswap_int(3,(int *)ibuf);
		}
		if( nprocs!=ibuf[2] )
		{
			SETERR("The number of PE=(%d) is different (in file PE=%d)\n",nprocs,ibuf[2]);
			return LIS_FAILS;
		}
		tbuf = (size_t *)lis_malloc( (nprocs+1)*sizeof(size_t) );
		if( tbuf==NULL )
		{
			SETERR("out of memory\n");
			return LIS_OUT_OF_MEMORY;
		}
		if( fread(tbuf, sizeof(size_t), nprocs+1, file)!=nprocs+1 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( mode )
		{
			lis_bswap_size_t(nprocs+1,(size_t *)tbuf);
		}
		fseek(file,tbuf[my_rank],SEEK_SET);

		if( fread(&ibuf[2], sizeof(int), 2, file)!=2 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( mode )
		{
			lis_bswap_int(2,(int *)ibuf);
		}
		lis_free(tbuf);
	}

	n   = ibuf[2];
	nnd = ibuf[3];
	err = lis_matrix_create(n,0,comm,A);
	if( err )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	col = (int *)lis_malloc( nprocs*sizeof(int) );
	if( col==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nprocs*nnd*sizeof(int) );
	if( index==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( n*nnd*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	if( ibuf[0] )
	{
		err = lis_vector_duplicateM(*A,b);
		if( err )
		{
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
	}
	if( ibuf[1] )
	{
		err = lis_vector_duplicateM(*A,x);
		if( err )
		{
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
	}

	if( mode==LIS_FMT_LIS_ASCII )
	{
		// nnd
		err = lis_fscan_int(nprocs,file,col,0);
		if( err )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// index
		err = lis_fscan_int(nprocs*nnd,file,index,0);
		if( err )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// value
		err = lis_fscan_double(n*nnd,file,value);
		if( err )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// RHS
		if( ibuf[0] )
		{
			err = lis_fscan_double(n,file,(*b)->value);
			if( err )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
		}
		// solution
		if( ibuf[1] )
		{
			err = lis_fscan_double(n,file,(*x)->value);
			if( err )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
		}
	}
	else
	{
		// nnd
		if( fread(col, sizeof(int), nprocs, file)!=nprocs )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// index
		if( fread(index, sizeof(int), nprocs*nnd, file)!=nprocs*nnd )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// value
		if( fread(value, sizeof(LIS_SCALAR), n*nnd, file)!=n*nnd )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// RHS
		if( ibuf[0] )
		{
			if( fread((*b)->value, sizeof(LIS_SCALAR), n, file)!=n )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
		}
		// solution
		if( ibuf[1] )
		{
			if( fread((*x)->value, sizeof(LIS_SCALAR), n, file)!=n )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
		}

		if( mode )
		{
			lis_bswap_int(nprocs,col);
			lis_bswap_int(nprocs*nnd,index);
			lis_bswap_double(n*nnd,value);
			if( ibuf[0] ) lis_bswap_double(n,(*b)->value);
			if( ibuf[1] ) lis_bswap_double(n,(*x)->value);
		}
	}

	err = lis_matrix_set_dia(col,index,value,*A);
	if( err ) return err;
	err = lis_matrix_assemble(A,LIS_MATRIX_DIA);
	if( err ) return err;

	return LIS_SUCCESS;
#else
	int				n,nnd,nprocs,my_rank,pe;
	int				i,j;
	int				err;
	int				ibuf[10];
	int				*index;
	char			cbuf[BUFSIZE];
	char			c;
	char			*p;
	size_t			*tbuf;
	LIS_SCALAR		*value;

#ifdef _MPI
	MPI_Comm_size(comm,&nprocs);
	MPI_Comm_rank(comm,&my_rank);
#else
	nprocs  = 1;
	my_rank = 0;
#endif
	if( mode==LIS_FMT_LIS_ASCII )
	{
		if( fgets(cbuf, BUFSIZE, file) == NULL )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( sscanf(cbuf, "%d %d %d",&ibuf[0], &ibuf[1], &ibuf[2])!=3 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( nprocs!=ibuf[2] )
		{
			SETERR("The number of PE=(%d) is different (in file PE=%d)\n",nprocs,ibuf[2]);
			return LIS_FAILS;
		}

		pe=-1;
		do
		{
			if( fgets(cbuf, BUFSIZE, file) == NULL )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
			if( cbuf[0]=='#' )
			{
				if( sscanf(cbuf, "%c %d %d %d",&c, &pe, &ibuf[2], &ibuf[3])!=4 )
				{
					SETERR("file i/o error\n");
					return LIS_FAILS;
				}
			}
		}while(pe!=my_rank);
	}
	else
	{
		if( fread(ibuf, sizeof(int), 3, file)!=3 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( mode )
		{
			lis_bswap_int(3,(int *)ibuf);
		}
		if( nprocs!=ibuf[2] )
		{
			SETERR("The number of PE=(%d) is different (in file PE=%d)\n",nprocs,ibuf[2]);
			return LIS_FAILS;
		}
		tbuf = (size_t *)lis_malloc( (nprocs+1)*sizeof(size_t) );
		if( tbuf==NULL )
		{
			SETERR("out of memory\n");
			return LIS_OUT_OF_MEMORY;
		}
		if( fread(tbuf, sizeof(size_t), nprocs+1, file)!=nprocs+1 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( mode )
		{
			lis_bswap_size_t(nprocs+1,(size_t *)tbuf);
		}
		fseek(file,tbuf[my_rank],SEEK_SET);

		if( fread(&ibuf[2], sizeof(int), 2, file)!=2 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( mode )
		{
			lis_bswap_int(2,(int *)ibuf);
		}
		lis_free(tbuf);
	}

	n   = ibuf[2];
	nnd = ibuf[3];
	err = lis_matrix_create(n,0,comm,A);
	if( err )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnd*sizeof(int) );
	if( index==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( n*nnd*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	if( ibuf[0] )
	{
		err = lis_vector_duplicateM(*A,b);
		if( err )
		{
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
	}
	if( ibuf[1] )
	{
		err = lis_vector_duplicateM(*A,x);
		if( err )
		{
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
	}

	if( mode==LIS_FMT_LIS_ASCII )
	{
		// index
		err = lis_fscan_int(nnd,file,index,0);
		if( err )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// value
		err = lis_fscan_double(n*nnd,file,value);
		if( err )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// RHS
		if( ibuf[0] )
		{
			err = lis_fscan_double(n,file,(*b)->value);
			if( err )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
		}
		// solution
		if( ibuf[1] )
		{
			err = lis_fscan_double(n,file,(*x)->value);
			if( err )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
		}
	}
	else
	{
		// index
		if( fread(index, sizeof(int), nnd, file)!=nnd )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// value
		if( fread(value, sizeof(LIS_SCALAR), n*nnd, file)!=n*nnd )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// RHS
		if( ibuf[0] )
		{
			if( fread((*b)->value, sizeof(LIS_SCALAR), n, file)!=n )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
		}
		// solution
		if( ibuf[1] )
		{
			if( fread((*x)->value, sizeof(LIS_SCALAR), n, file)!=n )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
		}

		if( mode )
		{
			lis_bswap_int(nnd,index);
			lis_bswap_double(n*nnd,value);
			if( ibuf[0] ) lis_bswap_double(n,(*b)->value);
			if( ibuf[1] ) lis_bswap_double(n,(*x)->value);
		}
	}

	err = lis_matrix_set_dia(nnd,index,value,*A);
	if( err ) return err;
	err = lis_matrix_assemble(A,LIS_MATRIX_DIA);
	if( err ) return err;

	return LIS_SUCCESS;
#endif
}

int lis_input_lis_ell(LIS_MATRIX *A, LIS_VECTOR *b, LIS_VECTOR *x, int mode, FILE *file, LIS_Comm comm)
{
	int				n,maxnzr,nprocs,my_rank,pe;
	int				i;
	int				err;
	int				ibuf[10];
	int				*index;
	char			cbuf[BUFSIZE];
	char			c;
	char			*p;
	size_t			*tbuf;
	LIS_SCALAR		*value;

#ifdef _MPI
	MPI_Comm_size(comm,&nprocs);
	MPI_Comm_rank(comm,&my_rank);
#else
	nprocs  = 1;
	my_rank = 0;
#endif
	if( mode==LIS_FMT_LIS_ASCII )
	{
		if( fgets(cbuf, BUFSIZE, file) == NULL )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( sscanf(cbuf, "%d %d %d",&ibuf[0], &ibuf[1], &ibuf[2])!=3 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( nprocs!=ibuf[2] )
		{
			SETERR("The number of PE=(%d) is different (in file PE=%d)\n",nprocs,ibuf[2]);
			return LIS_FAILS;
		}

		pe=-1;
		do
		{
			if( fgets(cbuf, BUFSIZE, file) == NULL )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
			if( cbuf[0]=='#' )
			{
				if( sscanf(cbuf, "%c %d %d %d",&c, &pe, &ibuf[2], &ibuf[3])!=4 )
				{
					SETERR("file i/o error\n");
					return LIS_FAILS;
				}
			}
		}while(pe!=my_rank);
	}
	else
	{
		if( fread(ibuf, sizeof(int), 3, file)!=3 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( mode )
		{
			lis_bswap_int(3,(int *)ibuf);
		}
		if( nprocs!=ibuf[2] )
		{
			SETERR("The number of PE=(%d) is different (in file PE=%d)\n",nprocs,ibuf[2]);
			return LIS_FAILS;
		}
		tbuf = (size_t *)lis_malloc( (nprocs+1)*sizeof(size_t) );
		if( tbuf==NULL )
		{
			SETERR("out of memory\n");
			return LIS_OUT_OF_MEMORY;
		}
		if( fread(tbuf, sizeof(size_t), nprocs+1, file)!=nprocs+1 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( mode )
		{
			lis_bswap_size_t(nprocs+1,(size_t *)tbuf);
		}
		fseek(file,tbuf[my_rank],SEEK_SET);

		if( fread(&ibuf[2], sizeof(int), 2, file)!=2 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( mode )
		{
			lis_bswap_int(2,(int *)ibuf);
		}
		lis_free(tbuf);
	}

	n      = ibuf[2];
	maxnzr = ibuf[3];
	err = lis_matrix_create(n,0,comm,A);
	if( err )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( n*maxnzr*sizeof(int) );
	if( index==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( n*maxnzr*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	if( ibuf[0] )
	{
		err = lis_vector_duplicateM(*A,b);
		if( err )
		{
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
	}
	if( ibuf[1] )
	{
		err = lis_vector_duplicateM(*A,x);
		if( err )
		{
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
	}

	if( mode==LIS_FMT_LIS_ASCII )
	{
		// index
		err = lis_fscan_int(n*maxnzr,file,index,0);
		if( err )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// value
		err = lis_fscan_double(n*maxnzr,file,value);
		if( err )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// RHS
		if( ibuf[0] )
		{
			err = lis_fscan_double(n,file,(*b)->value);
			if( err )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
		}
		// solution
		if( ibuf[1] )
		{
			err = lis_fscan_double(n,file,(*x)->value);
			if( err )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
		}
	}
	else
	{
		// index
		if( fread(index, sizeof(int), n*maxnzr, file)!=n*maxnzr )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// value
		if( fread(value, sizeof(LIS_SCALAR), n*maxnzr, file)!=n*maxnzr )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// RHS
		if( ibuf[0] )
		{
			if( fread((*b)->value, sizeof(LIS_SCALAR), n, file)!=n )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
		}
		// solution
		if( ibuf[1] )
		{
			if( fread((*x)->value, sizeof(LIS_SCALAR), n, file)!=n )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
		}

		if( mode )
		{
			lis_bswap_int(n*maxnzr,index);
			lis_bswap_double(n*maxnzr,value);
			if( ibuf[0] ) lis_bswap_double(n,(*b)->value);
			if( ibuf[1] ) lis_bswap_double(n,(*x)->value);
		}
	}

	err = lis_matrix_set_ell(maxnzr,index,value,*A);
	if( err ) return err;
	err = lis_matrix_assemble(A,LIS_MATRIX_ELL);
	if( err ) return err;

	return LIS_SUCCESS;
}

int lis_input_lis_jds(LIS_MATRIX *A, LIS_VECTOR *b, LIS_VECTOR *x, int mode, FILE *file, LIS_Comm comm)
{
#ifdef _OPENMP
	int				n,nnz,maxmaxnzr,nprocs,my_rank,pe;
	int				i;
	int				err;
	int				ibuf[10];
	int				*row,*ptr,*index,*maxnzr;
	char			cbuf[BUFSIZE];
	char			c;
	char			*p;
	size_t			*tbuf;
	LIS_SCALAR		*value;

	nprocs  = omp_get_max_threads();
	my_rank = 0;
	if( mode==LIS_FMT_LIS_ASCII )
	{
		if( fgets(cbuf, BUFSIZE, file) == NULL )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( sscanf(cbuf, "%d %d %d",&ibuf[0], &ibuf[1], &ibuf[2])!=3 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( nprocs!=ibuf[2] )
		{
			SETERR("The number of PE=(%d) is different (in file PE=%d)\n",nprocs,ibuf[2]);
			return LIS_FAILS;
		}

		pe=-1;
		do
		{
			if( fgets(cbuf, BUFSIZE, file) == NULL )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
			if( cbuf[0]=='#' )
			{
				if( sscanf(cbuf, "%c %d %d %d %d",&c, &pe, &ibuf[2], &ibuf[3], &ibuf[4])!=5 )
				{
					SETERR("file i/o error\n");
					return LIS_FAILS;
				}
			}
		}while(pe!=my_rank);
	}
	else
	{
		if( fread(ibuf, sizeof(int), 3, file)!=3 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( mode )
		{
			lis_bswap_int(3,(int *)ibuf);
		}
		if( nprocs!=ibuf[2] )
		{
			SETERR("The number of PE=(%d) is different (in file PE=%d)\n",nprocs,ibuf[2]);
			return LIS_FAILS;
		}
		tbuf = (size_t *)lis_malloc( 2*sizeof(size_t) );
		if( tbuf==NULL )
		{
			SETERR("out of memory\n");
			return LIS_OUT_OF_MEMORY;
		}
		if( fread(tbuf, sizeof(size_t), 2, file)!=2 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( mode )
		{
			lis_bswap_size_t(2,(size_t *)tbuf);
		}
		fseek(file,tbuf[my_rank],SEEK_SET);

		if( fread(&ibuf[2], sizeof(int), 3, file)!=3 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( mode )
		{
			lis_bswap_int(3,(int *)ibuf);
		}
		lis_free(tbuf);
	}

	n         = ibuf[2];
	nnz       = ibuf[3];
	maxmaxnzr = ibuf[4];
	err = lis_matrix_create(n,0,comm,A);
	if( err )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	maxnzr = (int *)lis_malloc( nprocs*sizeof(int) );
	if( maxnzr==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	row = (int *)lis_malloc( n*sizeof(int) );
	if( row==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	ptr = (int *)lis_malloc( nprocs*(maxmaxnzr+1)*sizeof(int) );
	if( ptr==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	if( ibuf[0] )
	{
		err = lis_vector_duplicateM(*A,b);
		if( err )
		{
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
	}
	if( ibuf[1] )
	{
		err = lis_vector_duplicateM(*A,x);
		if( err )
		{
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
	}

	if( mode==LIS_FMT_LIS_ASCII )
	{
		// maxnzr
		err = lis_fscan_int(nprocs,file,maxnzr,0);
		if( err )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// row
		err = lis_fscan_int(n,file,row,0);
		if( err )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// ptr
		err = lis_fscan_int(nprocs*(maxmaxnzr+1),file,ptr,0);
		if( err )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// index
		err = lis_fscan_int(nnz,file,index,0);
		if( err )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// value
		err = lis_fscan_double(nnz,file,value);
		if( err )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// RHS
		if( ibuf[0] )
		{
			err = lis_fscan_double(n,file,(*b)->value);
			if( err )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
		}
		// solution
		if( ibuf[1] )
		{
			err = lis_fscan_double(n,file,(*x)->value);
			if( err )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
		}
	}
	else
	{
		// maxnzr
		if( fread(maxnzr, sizeof(int), nprocs, file)!=nprocs )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// row
		if( fread(row, sizeof(int), n, file)!=n )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// ptr
		if( fread(ptr, sizeof(int), nprocs*(maxmaxnzr+1), file)!=nprocs*(maxmaxnzr+1) )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// index
		if( fread(index, sizeof(int), nnz, file)!=nnz )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// value
		if( fread(value, sizeof(LIS_SCALAR), nnz, file)!=nnz )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// RHS
		if( ibuf[0] )
		{
			if( fread((*b)->value, sizeof(LIS_SCALAR), n, file)!=n )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
		}
		// solution
		if( ibuf[1] )
		{
			if( fread((*x)->value, sizeof(LIS_SCALAR), n, file)!=n )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
		}

		if( mode )
		{
			lis_bswap_int(nprocs,maxnzr);
			lis_bswap_int(n,row);
			lis_bswap_int(nprocs*(maxmaxnzr+1),ptr);
			lis_bswap_int(nnz,index);
			lis_bswap_double(nnz,value);
			if( ibuf[0] ) lis_bswap_double(n,(*b)->value);
			if( ibuf[1] ) lis_bswap_double(n,(*x)->value);
		}
	}

	err = lis_matrix_set_jds(nnz,maxnzr,row,ptr,index,value,*A);
	if( err ) return err;
	err = lis_matrix_assemble(A,LIS_MATRIX_JDS);
	if( err ) return err;

	return LIS_SUCCESS;
#else
	int				n,nnz,maxnzr,nprocs,my_rank,pe;
	int				i;
	int				err;
	int				ibuf[10];
	int				*row,*ptr,*index;
	char			cbuf[BUFSIZE];
	char			c;
	char			*p;
	size_t			*tbuf;
	LIS_SCALAR		*value;

#ifdef _MPI
	MPI_Comm_size(comm,&nprocs);
	MPI_Comm_rank(comm,&my_rank);
#else
	nprocs  = 1;
	my_rank = 0;
#endif
	if( mode==LIS_FMT_LIS_ASCII )
	{
		if( fgets(cbuf, BUFSIZE, file) == NULL )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( sscanf(cbuf, "%d %d %d",&ibuf[0], &ibuf[1], &ibuf[2])!=3 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( nprocs!=ibuf[2] )
		{
			SETERR("The number of PE=(%d) is different (in file PE=%d)\n",nprocs,ibuf[2]);
			return LIS_FAILS;
		}

		pe=-1;
		do
		{
			if( fgets(cbuf, BUFSIZE, file) == NULL )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
			if( cbuf[0]=='#' )
			{
				if( sscanf(cbuf, "%c %d %d %d %d",&c, &pe, &ibuf[2], &ibuf[3], &ibuf[4])!=5 )
				{
					SETERR("file i/o error\n");
					return LIS_FAILS;
				}
			}
		}while(pe!=my_rank);
	}
	else
	{
		if( fread(ibuf, sizeof(int), 3, file)!=3 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( mode )
		{
			lis_bswap_int(3,(int *)ibuf);
		}
		if( nprocs!=ibuf[2] )
		{
			SETERR("The number of PE=(%d) is different (in file PE=%d)\n",nprocs,ibuf[2]);
			return LIS_FAILS;
		}
		tbuf = (size_t *)lis_malloc( (nprocs+1)*sizeof(size_t) );
		if( tbuf==NULL )
		{
			SETERR("out of memory\n");
			return LIS_OUT_OF_MEMORY;
		}
		if( fread(tbuf, sizeof(size_t), nprocs+1, file)!=nprocs+1 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( mode )
		{
			lis_bswap_size_t(nprocs+1,(size_t *)tbuf);
		}
		fseek(file,tbuf[my_rank],SEEK_SET);

		if( fread(&ibuf[2], sizeof(int), 3, file)!=3 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( mode )
		{
			lis_bswap_int(3,(int *)ibuf);
		}
		lis_free(tbuf);
	}

	n      = ibuf[2];
	nnz    = ibuf[3];
	maxnzr = ibuf[4];
	err = lis_matrix_create(n,0,comm,A);
	if( err )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	row = (int *)lis_malloc( n*sizeof(int) );
	if( row==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	ptr = (int *)lis_malloc( (maxnzr+1)*sizeof(int) );
	if( ptr==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	if( ibuf[0] )
	{
		err = lis_vector_duplicateM(*A,b);
		if( err )
		{
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
	}
	if( ibuf[1] )
	{
		err = lis_vector_duplicateM(*A,x);
		if( err )
		{
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
	}

	if( mode==LIS_FMT_LIS_ASCII )
	{
		// row
		err = lis_fscan_int(n,file,row,0);
		if( err )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// ptr
		err = lis_fscan_int(maxnzr+1,file,ptr,0);
		if( err )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// index
		err = lis_fscan_int(nnz,file,index,0);
		if( err )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// value
		err = lis_fscan_double(nnz,file,value);
		if( err )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// RHS
		if( ibuf[0] )
		{
			err = lis_fscan_double(n,file,(*b)->value);
			if( err )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
		}
		// solution
		if( ibuf[1] )
		{
			err = lis_fscan_double(n,file,(*x)->value);
			if( err )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
		}
	}
	else
	{
		// row
		if( fread(row, sizeof(int), n, file)!=n )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// ptr
		if( fread(ptr, sizeof(int), maxnzr+1, file)!=maxnzr+1 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// index
		if( fread(index, sizeof(int), nnz, file)!=nnz )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// value
		if( fread(value, sizeof(LIS_SCALAR), nnz, file)!=nnz )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// RHS
		if( ibuf[0] )
		{
			if( fread((*b)->value, sizeof(LIS_SCALAR), n, file)!=n )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
		}
		// solution
		if( ibuf[1] )
		{
			if( fread((*x)->value, sizeof(LIS_SCALAR), n, file)!=n )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
		}

		if( mode )
		{
			lis_bswap_int(n,row);
			lis_bswap_int(maxnzr+1,ptr);
			lis_bswap_int(nnz,index);
			lis_bswap_double(nnz,value);
			if( ibuf[0] ) lis_bswap_double(n,(*b)->value);
			if( ibuf[1] ) lis_bswap_double(n,(*x)->value);
		}
	}

	err = lis_matrix_set_jds(nnz,maxnzr,row,ptr,index,value,*A);
	if( err ) return err;
	err = lis_matrix_assemble(A,LIS_MATRIX_JDS);
	if( err ) return err;

	return LIS_SUCCESS;
#endif
}

int lis_input_lis_bsr(LIS_MATRIX *A, LIS_VECTOR *b, LIS_VECTOR *x, int mode, FILE *file, LIS_Comm comm)
{
	int				n,nr,nc,bnr,bnc,bnnz,nprocs,my_rank,pe;
	int				i;
	int				err;
	int				ibuf[10];
	int				*bptr,*bindex;
	char			cbuf[BUFSIZE];
	char			c;
	char			*p;
	size_t			*tbuf;
	LIS_SCALAR		*value;

#ifdef _MPI
	MPI_Comm_size(comm,&nprocs);
	MPI_Comm_rank(comm,&my_rank);
#else
	nprocs  = 1;
	my_rank = 0;
#endif
	if( mode==LIS_FMT_LIS_ASCII )
	{
		if( fgets(cbuf, BUFSIZE, file) == NULL )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( sscanf(cbuf, "%d %d %d",&ibuf[0], &ibuf[1], &ibuf[2])!=3 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( nprocs!=ibuf[2] )
		{
			SETERR("The number of PE=(%d) is different (in file PE=%d)\n",nprocs,ibuf[2]);
			return LIS_FAILS;
		}

		pe=-1;
		do
		{
			if( fgets(cbuf, BUFSIZE, file) == NULL )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
			if( cbuf[0]=='#' )
			{
				if( sscanf(cbuf, "%c %d %d %d %d %d %d %d",&c, &pe, &ibuf[2], &ibuf[3], &ibuf[4], &ibuf[5], &ibuf[6], &ibuf[7])!=8 )
				{
					SETERR("file i/o error\n");
					return LIS_FAILS;
				}
			}
		}while(pe!=my_rank);
	}
	else
	{
		if( fread(ibuf, sizeof(int), 3, file)!=3 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( mode )
		{
			lis_bswap_int(3,(int *)ibuf);
		}
		if( nprocs!=ibuf[2] )
		{
			SETERR("The number of PE=(%d) is different (in file PE=%d)\n",nprocs,ibuf[2]);
			return LIS_FAILS;
		}
		tbuf = (size_t *)lis_malloc( (nprocs+1)*sizeof(size_t) );
		if( tbuf==NULL )
		{
			SETERR("out of memory\n");
			return LIS_OUT_OF_MEMORY;
		}
		if( fread(tbuf, sizeof(size_t), nprocs+1, file)!=nprocs+1 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( mode )
		{
			lis_bswap_size_t(nprocs+1,(size_t *)tbuf);
		}
		fseek(file,tbuf[my_rank],SEEK_SET);

		if( fread(&ibuf[2], sizeof(int), 6, file)!=6 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( mode )
		{
			lis_bswap_int(6,(int *)ibuf);
		}
		lis_free(tbuf);
	}

	n    = ibuf[2];
	nr   = ibuf[3];
	nc   = ibuf[4];
	bnr  = ibuf[5];
	bnc  = ibuf[6];
	bnnz = ibuf[7];
	err = lis_matrix_create(n,0,comm,A);
	if( err )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	bptr = (int *)lis_malloc( (nr+1)*sizeof(int) );
	if( bptr==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	bindex = (int *)lis_malloc( bnnz*sizeof(int) );
	if( bindex==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( bnnz*bnr*bnc*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	if( ibuf[0] )
	{
		err = lis_vector_duplicateM(*A,b);
		if( err )
		{
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
	}
	if( ibuf[1] )
	{
		err = lis_vector_duplicateM(*A,x);
		if( err )
		{
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
	}

	if( mode==LIS_FMT_LIS_ASCII )
	{
		// bptr
		err = lis_fscan_int(nr+1,file,bptr,0);
		if( err )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// bindex
		err = lis_fscan_int(bnnz,file,bindex,0);
		if( err )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// value
		err = lis_fscan_double(bnnz*bnr*bnc,file,value);
		if( err )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// RHS
		if( ibuf[0] )
		{
			err = lis_fscan_double(n,file,(*b)->value);
			if( err )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
		}
		// solution
		if( ibuf[1] )
		{
			err = lis_fscan_double(n,file,(*x)->value);
			if( err )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
		}
	}
	else
	{
		// bptr
		if( fread(bptr, sizeof(int), nr+1, file)!=nr+1 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// bindex
		if( fread(bindex, sizeof(int), bnnz, file)!=bnnz )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// value
		if( fread(value, sizeof(LIS_SCALAR), bnnz*bnr*bnc, file)!=bnnz*bnr*bnc )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// RHS
		if( ibuf[0] )
		{
			if( fread((*b)->value, sizeof(LIS_SCALAR), n, file)!=n )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
		}
		// solution
		if( ibuf[1] )
		{
			if( fread((*x)->value, sizeof(LIS_SCALAR), n, file)!=n )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
		}

		if( mode )
		{
			lis_bswap_int(nr+1,bptr);
			lis_bswap_int(bnnz,bindex);
			lis_bswap_double(bnnz*bnr*bnc,value);
			if( ibuf[0] ) lis_bswap_double(n,(*b)->value);
			if( ibuf[1] ) lis_bswap_double(n,(*x)->value);
		}
	}

	err = lis_matrix_set_bsr(bnr,bnc,bnnz,bptr,bindex,value,*A);
	if( err ) return err;
	err = lis_matrix_assemble(A,LIS_MATRIX_BSR);
	if( err ) return err;

	return LIS_SUCCESS;
}

int lis_input_lis_bsc(LIS_MATRIX *A, LIS_VECTOR *b, LIS_VECTOR *x, int mode, FILE *file, LIS_Comm comm)
{
	int				n,nr,nc,bnr,bnc,bnnz,nprocs,my_rank,pe;
	int				i;
	int				err;
	int				ibuf[10];
	int				*bptr,*bindex;
	char			cbuf[BUFSIZE];
	char			c;
	char			*p;
	size_t			*tbuf;
	LIS_SCALAR		*value;

#ifdef _MPI
	MPI_Comm_size(comm,&nprocs);
	MPI_Comm_rank(comm,&my_rank);
#else
	nprocs  = 1;
	my_rank = 0;
#endif
	if( mode==LIS_FMT_LIS_ASCII )
	{
		if( fgets(cbuf, BUFSIZE, file) == NULL )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( sscanf(cbuf, "%d %d %d",&ibuf[0], &ibuf[1], &ibuf[2])!=3 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( nprocs!=ibuf[2] )
		{
			SETERR("The number of PE=(%d) is different (in file PE=%d)\n",nprocs,ibuf[2]);
			return LIS_FAILS;
		}

		pe=-1;
		do
		{
			if( fgets(cbuf, BUFSIZE, file) == NULL )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
			if( cbuf[0]=='#' )
			{
				if( sscanf(cbuf, "%c %d %d %d %d %d %d %d",&c, &pe, &ibuf[2], &ibuf[3], &ibuf[4], &ibuf[5], &ibuf[6], &ibuf[7])!=8 )
				{
					SETERR("file i/o error\n");
					return LIS_FAILS;
				}
			}
		}while(pe!=my_rank);
	}
	else
	{
		if( fread(ibuf, sizeof(int), 3, file)!=3 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( mode )
		{
			lis_bswap_int(3,(int *)ibuf);
		}
		if( nprocs!=ibuf[2] )
		{
			SETERR("The number of PE=(%d) is different (in file PE=%d)\n",nprocs,ibuf[2]);
			return LIS_FAILS;
		}
		tbuf = (size_t *)lis_malloc( (nprocs+1)*sizeof(size_t) );
		if( tbuf==NULL )
		{
			SETERR("out of memory\n");
			return LIS_OUT_OF_MEMORY;
		}
		if( fread(tbuf, sizeof(size_t), nprocs+1, file)!=nprocs+1 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( mode )
		{
			lis_bswap_size_t(nprocs+1,(size_t *)tbuf);
		}
		fseek(file,tbuf[my_rank],SEEK_SET);

		if( fread(&ibuf[2], sizeof(int), 6, file)!=6 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( mode )
		{
			lis_bswap_int(6,(int *)ibuf);
		}
		lis_free(tbuf);
	}

	n    = ibuf[2];
	nr   = ibuf[3];
	nc   = ibuf[4];
	bnr  = ibuf[5];
	bnc  = ibuf[6];
	bnnz = ibuf[7];
	err = lis_matrix_create(n,0,comm,A);
	if( err )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	bptr = (int *)lis_malloc( (nc+1)*sizeof(int) );
	if( bptr==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	bindex = (int *)lis_malloc( bnnz*sizeof(int) );
	if( bindex==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( bnnz*bnr*bnc*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	if( ibuf[0] )
	{
		err = lis_vector_duplicateM(*A,b);
		if( err )
		{
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
	}
	if( ibuf[1] )
	{
		err = lis_vector_duplicateM(*A,x);
		if( err )
		{
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
	}

	if( mode==LIS_FMT_LIS_ASCII )
	{
		// bptr
		err = lis_fscan_int(nc+1,file,bptr,0);
		if( err )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// bindex
		err = lis_fscan_int(bnnz,file,bindex,0);
		if( err )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// value
		err = lis_fscan_double(bnnz*bnr*bnc,file,value);
		if( err )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// RHS
		if( ibuf[0] )
		{
			err = lis_fscan_double(n,file,(*b)->value);
			if( err )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
		}
		// solution
		if( ibuf[1] )
		{
			err = lis_fscan_double(n,file,(*x)->value);
			if( err )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
		}
	}
	else
	{
		// bptr
		if( fread(bptr, sizeof(int), nc+1, file)!=nc+1 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// bindex
		if( fread(bindex, sizeof(int), bnnz, file)!=bnnz )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// value
		if( fread(value, sizeof(LIS_SCALAR), bnnz*bnr*bnc, file)!=bnnz*bnr*bnc )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// RHS
		if( ibuf[0] )
		{
			if( fread((*b)->value, sizeof(LIS_SCALAR), n, file)!=n )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
		}
		// solution
		if( ibuf[1] )
		{
			if( fread((*x)->value, sizeof(LIS_SCALAR), n, file)!=n )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
		}

		if( mode )
		{
			lis_bswap_int(nc+1,bptr);
			lis_bswap_int(bnnz,bindex);
			lis_bswap_double(bnnz*bnr*bnc,value);
			if( ibuf[0] ) lis_bswap_double(n,(*b)->value);
			if( ibuf[1] ) lis_bswap_double(n,(*x)->value);
		}
	}

	err = lis_matrix_set_bsc(bnr,bnc,bnnz,bptr,bindex,value,*A);
	if( err ) return err;
	err = lis_matrix_assemble(A,LIS_MATRIX_BSC);
	if( err ) return err;

	return LIS_SUCCESS;
}

int lis_input_lis_vbr(LIS_MATRIX *A, LIS_VECTOR *b, LIS_VECTOR *x, int mode, FILE *file, LIS_Comm comm)
{
	int				n,nr,nc,nnz,bnnz,nprocs,my_rank,pe;
	int				i;
	int				err;
	int				ibuf[10];
	int				*row,*col,*ptr,*bptr,*bindex;
	char			cbuf[BUFSIZE];
	char			c;
	char			*p;
	size_t			*tbuf;
	LIS_SCALAR		*value;

#ifdef _MPI
	MPI_Comm_size(comm,&nprocs);
	MPI_Comm_rank(comm,&my_rank);
#else
	nprocs  = 1;
	my_rank = 0;
#endif
	if( mode==LIS_FMT_LIS_ASCII )
	{
		if( fgets(cbuf, BUFSIZE, file) == NULL )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( sscanf(cbuf, "%d %d %d",&ibuf[0], &ibuf[1], &ibuf[2])!=3 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( nprocs!=ibuf[2] )
		{
			SETERR("The number of PE=(%d) is different (in file PE=%d)\n",nprocs,ibuf[2]);
			return LIS_FAILS;
		}

		pe=-1;
		do
		{
			if( fgets(cbuf, BUFSIZE, file) == NULL )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
			if( cbuf[0]=='#' )
			{
				if( sscanf(cbuf, "%c %d %d %d %d %d %d",&c, &pe, &ibuf[2], &ibuf[3], &ibuf[4], &ibuf[5], &ibuf[6])!=7 )
				{
					SETERR("file i/o error\n");
					return LIS_FAILS;
				}
			}
		}while(pe!=my_rank);
	}
	else
	{
		if( fread(ibuf, sizeof(int), 3, file)!=3 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( mode )
		{
			lis_bswap_int(3,(int *)ibuf);
		}
		if( nprocs!=ibuf[2] )
		{
			SETERR("The number of PE=(%d) is different (in file PE=%d)\n",nprocs,ibuf[2]);
			return LIS_FAILS;
		}
		tbuf = (size_t *)lis_malloc( (nprocs+1)*sizeof(size_t) );
		if( tbuf==NULL )
		{
			SETERR("out of memory\n");
			return LIS_OUT_OF_MEMORY;
		}
		if( fread(tbuf, sizeof(size_t), nprocs+1, file)!=nprocs+1 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( mode )
		{
			lis_bswap_size_t(nprocs+1,(size_t *)tbuf);
		}
		fseek(file,tbuf[my_rank],SEEK_SET);

		if( fread(&ibuf[2], sizeof(int), 5, file)!=5 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( mode )
		{
			lis_bswap_int(5,(int *)ibuf);
		}
		lis_free(tbuf);
	}

	n    = ibuf[2];
	nnz  = ibuf[3];
	nr   = ibuf[4];
	nc   = ibuf[5];
	bnnz = ibuf[6];
	err = lis_matrix_create(n,0,comm,A);
	if( err )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	row = (int *)lis_malloc( (nr+1)*sizeof(int) );
	if( row==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	col = (int *)lis_malloc( (nc+1)*sizeof(int) );
	if( col==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	ptr = (int *)lis_malloc( (bnnz+1)*sizeof(int) );
	if( ptr==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	bptr = (int *)lis_malloc( (nr+1)*sizeof(int) );
	if( bptr==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	bindex = (int *)lis_malloc( bnnz*sizeof(int) );
	if( bindex==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	if( ibuf[0] )
	{
		err = lis_vector_duplicateM(*A,b);
		if( err )
		{
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
	}
	if( ibuf[1] )
	{
		err = lis_vector_duplicateM(*A,x);
		if( err )
		{
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
	}

	if( mode==LIS_FMT_LIS_ASCII )
	{
		// row
		err = lis_fscan_int(nr+1,file,row,0);
		if( err )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// col
		err = lis_fscan_int(nc+1,file,col,0);
		if( err )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// ptr
		err = lis_fscan_int(bnnz+1,file,ptr,0);
		if( err )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// bptr
		err = lis_fscan_int(nr+1,file,bptr,0);
		if( err )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// bindex
		err = lis_fscan_int(bnnz,file,bindex,0);
		if( err )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// value
		err = lis_fscan_double(nnz,file,value);
		if( err )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// RHS
		if( ibuf[0] )
		{
			err = lis_fscan_double(n,file,(*b)->value);
			if( err )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
		}
		// solution
		if( ibuf[1] )
		{
			err = lis_fscan_double(n,file,(*x)->value);
			if( err )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
		}
	}
	else
	{
		// row
		if( fread(row, sizeof(int), nr+1, file)!=nr+1 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// col
		if( fread(col, sizeof(int), nc+1, file)!=nc+1 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// ptr
		if( fread(ptr, sizeof(int), bnnz+1, file)!=bnnz+1 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// bptr
		if( fread(bptr, sizeof(int), nr+1, file)!=nr+1 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// bindex
		if( fread(bindex, sizeof(int), bnnz, file)!=bnnz )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// value
		if( fread(value, sizeof(LIS_SCALAR), nnz, file)!=nnz )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// RHS
		if( ibuf[0] )
		{
			if( fread((*b)->value, sizeof(LIS_SCALAR), n, file)!=n )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
		}
		// solution
		if( ibuf[1] )
		{
			if( fread((*x)->value, sizeof(LIS_SCALAR), n, file)!=n )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
		}

		if( mode )
		{
			lis_bswap_int(nr+1,row);
			lis_bswap_int(nc+1,col);
			lis_bswap_int(bnnz+1,ptr);
			lis_bswap_int(nr+1,bptr);
			lis_bswap_int(bnnz,bindex);
			lis_bswap_double(nnz,value);
			if( ibuf[0] ) lis_bswap_double(n,(*b)->value);
			if( ibuf[1] ) lis_bswap_double(n,(*x)->value);
		}
	}

	err = lis_matrix_set_vbr(nnz,nr,nc,bnnz,row,col,ptr,bptr,bindex,value,*A);
	if( err ) return err;
	err = lis_matrix_assemble(A,LIS_MATRIX_VBR);
	if( err ) return err;

	return LIS_SUCCESS;
}

int lis_input_lis_coo(LIS_MATRIX *A, LIS_VECTOR *b, LIS_VECTOR *x, int mode, FILE *file, LIS_Comm comm)
{
	int				n,nnz,nprocs,my_rank,pe;
	int				i;
	int				err;
	int				ibuf[10];
	int				*row,*col;
	char			cbuf[BUFSIZE];
	char			c;
	char			*p;
	size_t			*tbuf;
	LIS_SCALAR		*value;

#ifdef _MPI
	MPI_Comm_size(comm,&nprocs);
	MPI_Comm_rank(comm,&my_rank);
#else
	nprocs  = 1;
	my_rank = 0;
#endif
	if( mode==LIS_FMT_LIS_ASCII )
	{
		if( fgets(cbuf, BUFSIZE, file) == NULL )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( sscanf(cbuf, "%d %d %d",&ibuf[0], &ibuf[1], &ibuf[2])!=3 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( nprocs!=ibuf[2] )
		{
			SETERR("The number of PE=(%d) is different (in file PE=%d)\n",nprocs,ibuf[2]);
			return LIS_FAILS;
		}
		pe=-1;
		do
		{
			if( fgets(cbuf, BUFSIZE, file) == NULL )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
			if( cbuf[0]=='#' )
			{
				if( sscanf(cbuf, "%c %d %d %d",&c, &pe, &ibuf[2], &ibuf[3])!=4 )
				{
					SETERR("file i/o error\n");
					return LIS_FAILS;
				}
			}
		}while(pe!=my_rank);
	}
	else
	{
		if( fread(ibuf, sizeof(int), 3, file)!=3 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( mode )
		{
			lis_bswap_int(3,(int *)ibuf);
		}
		if( nprocs!=ibuf[2] )
		{
			SETERR("The number of PE=(%d) is different (in file PE=%d)\n",nprocs,ibuf[2]);
			return LIS_FAILS;
		}
		tbuf = (size_t *)lis_malloc( (nprocs+1)*sizeof(size_t) );
		if( tbuf==NULL )
		{
			SETERR("out of memory\n");
			return LIS_OUT_OF_MEMORY;
		}
		if( fread(tbuf, sizeof(size_t), nprocs+1, file)!=nprocs+1 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( mode )
		{
			lis_bswap_size_t(nprocs+1,(size_t *)tbuf);
		}
		fseek(file,tbuf[my_rank],SEEK_SET);

		if( fread(&ibuf[2], sizeof(int), 2, file)!=2 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( mode )
		{
			lis_bswap_int(2,(int *)ibuf);
		}
		lis_free(tbuf);
	}

	n   = ibuf[2];
	nnz = ibuf[3];
	err = lis_matrix_create(n,0,comm,A);
	if( err )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	row = (int *)lis_malloc( nnz*sizeof(int) );
	if( row==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	col = (int *)lis_malloc( nnz*sizeof(int) );
	if( col==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	if( ibuf[0] )
	{
		err = lis_vector_duplicateM(*A,b);
		if( err )
		{
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
	}
	if( ibuf[1] )
	{
		err = lis_vector_duplicateM(*A,x);
		if( err )
		{
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
	}

	if( mode==LIS_FMT_LIS_ASCII )
	{
		// row
		err = lis_fscan_int(nnz,file,row,0);
		if( err )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// col
		err = lis_fscan_int(nnz,file,col,0);
		if( err )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// value
		err = lis_fscan_double(nnz,file,value);
		if( err )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// RHS
		if( ibuf[0] )
		{
			err = lis_fscan_double(n,file,(*b)->value);
			if( err )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
		}
		// solution
		if( ibuf[1] )
		{
			err = lis_fscan_double(n,file,(*x)->value);
			if( err )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
		}
	}
	else
	{
		// row
		if( fread(row, sizeof(int), nnz, file)!=nnz )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// col
		if( fread(col, sizeof(int), nnz, file)!=nnz )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// value
		if( fread(value, sizeof(LIS_SCALAR), nnz, file)!=nnz )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// RHS
		if( ibuf[0] )
		{
			if( fread((*b)->value, sizeof(LIS_SCALAR), n, file)!=n )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
		}
		// solution
		if( ibuf[1] )
		{
			if( fread((*x)->value, sizeof(LIS_SCALAR), n, file)!=n )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
		}

		if( mode )
		{
			lis_bswap_int(nnz,row);
			lis_bswap_int(nnz,col);
			lis_bswap_double(nnz,value);
			if( ibuf[0] ) lis_bswap_double(n,(*b)->value);
			if( ibuf[1] ) lis_bswap_double(n,(*x)->value);
		}
	}

	err = lis_matrix_set_coo(nnz,row,col,value,*A);
	if( err ) return err;
	err = lis_matrix_assemble(A,LIS_MATRIX_COO);
	if( err ) return err;

	return LIS_SUCCESS;
}

int lis_input_lis_dns(LIS_MATRIX *A, LIS_VECTOR *b, LIS_VECTOR *x, int mode, FILE *file, LIS_Comm comm)
{
	int				n,gn,nprocs,my_rank,pe;
	int				i;
	int				err;
	int				ibuf[10];
	char			cbuf[BUFSIZE];
	char			c;
	char			*p;
	size_t			*tbuf;
	LIS_SCALAR		*value;

#ifdef _MPI
	MPI_Comm_size(comm,&nprocs);
	MPI_Comm_rank(comm,&my_rank);
#else
	nprocs  = 1;
	my_rank = 0;
#endif
	if( mode==LIS_FMT_LIS_ASCII )
	{
		if( fgets(cbuf, BUFSIZE, file) == NULL )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( sscanf(cbuf, "%d %d %d",&ibuf[0], &ibuf[1], &ibuf[2])!=3 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( nprocs!=ibuf[2] )
		{
			SETERR("The number of PE=(%d) is different (in file PE=%d)\n",nprocs,ibuf[2]);
			return LIS_FAILS;
		}
		pe=-1;
		do
		{
			if( fgets(cbuf, BUFSIZE, file) == NULL )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
			if( cbuf[0]=='#' )
			{
				if( sscanf(cbuf, "%c %d %d",&c, &pe, &ibuf[2])!=3 )
				{
					SETERR("file i/o error\n");
					return LIS_FAILS;
				}
			}
		}while(pe!=my_rank);
	}
	else
	{
		if( fread(ibuf, sizeof(int), 3, file)!=3 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( mode )
		{
			lis_bswap_int(3,(int *)ibuf);
		}
		if( nprocs!=ibuf[2] )
		{
			SETERR("The number of PE=(%d) is different (in file PE=%d)\n",nprocs,ibuf[2]);
			return LIS_FAILS;
		}
		tbuf = (size_t *)lis_malloc( (nprocs+1)*sizeof(size_t) );
		if( tbuf==NULL )
		{
			SETERR("out of memory\n");
			return LIS_OUT_OF_MEMORY;
		}
		if( fread(tbuf, sizeof(size_t), nprocs+1, file)!=nprocs+1 )
		{
			lis_free(tbuf);
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( mode )
		{
			lis_bswap_size_t(nprocs+1,(size_t *)tbuf);
		}
		fseek(file,tbuf[my_rank],SEEK_SET);

		if( fread(&ibuf[2], sizeof(int), 1, file)!=1 )
		{
			lis_free(tbuf);
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( mode )
		{
			lis_bswap_int(1,(int *)ibuf);
		}
		lis_free(tbuf);
	}

	n   = ibuf[2];
	err = lis_matrix_create(n,0,comm,A);
	if( err )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	gn = (*A)->gn;
	value = (LIS_SCALAR *)lis_malloc( gn*n*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	if( ibuf[0] )
	{
		err = lis_vector_duplicateM(*A,b);
		if( err )
		{
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
	}
	if( ibuf[1] )
	{
		err = lis_vector_duplicateM(*A,x);
		if( err )
		{
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
	}

	if( mode==LIS_FMT_LIS_ASCII )
	{
		// value
		err = lis_fscan_double(gn*n,file,value);
		if( err )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// RHS
		if( ibuf[0] )
		{
			err = lis_fscan_double(n,file,(*b)->value);
			if( err )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
		}
		// solution
		if( ibuf[1] )
		{
			err = lis_fscan_double(n,file,(*x)->value);
			if( err )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
		}
	}
	else
	{
		// value
		if( fread(value, sizeof(LIS_SCALAR), gn*n, file)!=gn*n )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		// RHS
		if( ibuf[0] )
		{
			if( fread((*b)->value, sizeof(LIS_SCALAR), n, file)!=n )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
		}
		// solution
		if( ibuf[1] )
		{
			if( fread((*x)->value, sizeof(LIS_SCALAR), n, file)!=n )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
		}

		if( mode )
		{
			lis_bswap_double(gn*n,value);
			if( ibuf[0] ) lis_bswap_double(n,(*b)->value);
			if( ibuf[1] ) lis_bswap_double(n,(*x)->value);
		}
	}

	err = lis_matrix_set_dns(value,*A);
	if( err ) return err;
	err = lis_matrix_assemble(A,LIS_MATRIX_DNS);
	if( err ) return err;

	return LIS_SUCCESS;
}
