#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <math.h>
#include <string.h>
#ifdef _OPENMP
	#include <omp.h>
#endif
#include "lis.h"
#include "lis_precon.h"

int lis_precon_create_is(LIS_PRECON *precon, LIS_VECTOR bb)
{
#ifdef _OPENMP
	int i,j,k,ik,m,id,threads;
	int n,nnz,ja,jj,jb,jcol,jpos,err;
	int *iw,*iw2,*index;
	int *diag;
	int *nnzs;
	int *ptr;
	int is,ie,n1,n2;
	double ctimes;
	LIS_VECTOR Pbb,dd;
	LIS_SCALAR val,t;
	LIS_SCALAR w;
	LIS_SCALAR *b,*Pb,*d;
	LIS_SCALAR *value;
	LIS_MATRIX A, B, C;
	LIS_MATRIX_PARAMS paramm;

	err = lis_vector_duplicate(bb,&Pbb);
	if( err )
	{
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

	err = lis_vector_duplicate(bb,&dd);
	if( err )
	{
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

	A    = (*precon)->Amat;
	B    = (*precon)->Amat;
	diag = (*precon)->diag;
	b    = bb->value;
	Pb   = Pbb->value;
	d    = dd->value;
	n    = A->n;
	w    = (*precon)->params[LIS_PARAMS_ALPHA]; 
	m    = (*precon)->options[LIS_OPTIONS_M] + 1; 


#define PADD 64
	threads = omp_get_max_threads();
	nnzs  = (int *)lis_malloc(threads*PADD*sizeof(int));
	ptr   = (int *)lis_malloc((n+1)*sizeof(int));
	lis_matrix_create(n,0,A->comm,&C);

#pragma omp parallel for private(i,j)
	for(i=0; i<n; i++)
	{
		lis_sort_id(A->ptr[i],A->ptr[i+1]-1,A->index,A->value);
		for(j=A->ptr[i];j<A->ptr[i+1];j++)
		{
			if( i==A->index[j] )
			{
				diag[i] = j;
				break;
			}
		}
	}
#pragma omp parallel private(i,ja,k,jj,jb,jcol,jpos,j,iw,iw2,id,is,ie,n1,n2)
{
	id = omp_get_thread_num();
	n1 = n/threads;
	n2 = n%threads;
	is = id<n2 ? (n1+1)*id : n1*id+n2;
	ie = id<n2 ? is+n1+1   : is+n1;
	is = (n1+1)*id;
	ie = is+n1+1;
	iw = lis_malloc( 2*n*sizeof(int) );
//	if( iw==NULL ) return LIS_OUT_OF_MEMORY;
	memset( iw,0,n*sizeof(int) );
	iw2 = iw + n;
	nnzs[id*PADD] = 0;
#pragma omp for 
	for(i=0;i<n;i++)
	{
		k = 0;
		n2 = _min(diag[i]+m,A->ptr[i+1]);
		for(ja=diag[i];ja<n2;ja++)
		{
			jj = A->index[ja];
//			if( jj>=ie ) break;
			for(jb=B->ptr[jj];jb<B->ptr[jj+1];jb++)
			{
				jcol = B->index[jb];
				jpos = iw[jcol];
				if( jpos==0 )
				{
					iw2[k] = jcol;
					k++;
					nnzs[id*PADD]++;
					iw[jcol] = k;
				}
			}
		}
		ptr[i+1] = k;
		for(j=0;j<k;j++)
		{
			iw[iw2[j]] = 0;
		}
	}
	lis_free(iw);
}
	nnz = 0;
	for(i=0;i<threads;i++)
	{
		nnz += nnzs[i*PADD];
	}

	ptr[0] = 0;
	for(i=0;i<n;i++)
	{
		ptr[i+1] += ptr[i]; 
	}

	index  = (int *)lis_malloc(nnz*sizeof(int));
	value  = (LIS_SCALAR *)lis_malloc(nnz*sizeof(LIS_SCALAR));


#pragma omp parallel private(i,ja,k,jj,jb,jcol,jpos,j,iw,id,val,t,is,ie,n1,n2)
{
	id = omp_get_thread_num();
	n1 = n/threads;
	n2 = n%threads;
	is = id>n2 ? (n1+1)*id : n1*id+n2;
	ie = id>n2 ? is+n1+1   : is+n1;
	is = (n1+1)*id;
	ie = is+n1+1;
	iw = lis_malloc( n*sizeof(int) );
//	if( iw==NULL ) return LIS_OUT_OF_MEMORY;
	memset( iw,0,n*sizeof(int) );
#pragma omp for
	for(i=0;i<n;i++)
	{
		k = ptr[i];
		t = 0.0;
		for(ja=diag[i];ja<_min(diag[i]+m,A->ptr[i+1]);ja++)
		{
			jj = A->index[ja];
			if( jj>=ie ) break;
			val = A->value[ja];
			if( jj!=i ) val = -w * val;
			for(jb=B->ptr[jj];jb<B->ptr[jj+1];jb++)
			{
				jcol = B->index[jb];
				jpos = iw[jcol];
				if( jpos==0 )
				{
					index[k]   = jcol;
					value[k]   = val * B->value[jb];
					k++;
					iw[jcol] = k;
				}
				else
				{
					value[jpos-1] += val * B->value[jb];
				}
			}
			t += val * b[jj];
		}
		for(j=ptr[i];j<k;j++)
		{
			iw[index[j]] = 0;
		}
		Pb[i] = t;
	}
	lis_free(iw);
}
	lis_matrix_set_crs(nnz,ptr,index,value,C);
	lis_matrix_assemble(&C,LIS_MATRIX_CRS);

	lis_matrix_scaling(C,Pbb,dd,LIS_SCALE_JACOBI);

	(*precon)->Pmat = C;
	(*precon)->Pb   = Pbb;
	lis_free(iw);
	lis_vector_destroy(dd);
#else
	int i,j,k,m;
	int n,gn,nnz,ja,jj,jb,jcol,jpos,err;
	int *iw,*iw2,*index;
	int *diag;
	int *ptr;
	int n2;
	LIS_VECTOR Pbb,dd;
	LIS_SCALAR val,t;
	LIS_SCALAR w;
	LIS_SCALAR *b,*Pb,*d;
	LIS_SCALAR *value;
	LIS_MATRIX A, B, C;

	err = lis_vector_duplicate(bb,&Pbb);
	if( err )
	{
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

	err = lis_vector_duplicate(bb,&dd);
	if( err )
	{
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

	A    = (*precon)->Amat;
	B    = (*precon)->Amat;
	diag = (*precon)->diag;
	b    = bb->value;
	Pb   = Pbb->value;
	d    = dd->value;
	n    = A->n;
	gn   = A->gn;
	w    = (*precon)->params[LIS_PARAMS_ALPHA]; 
	m    = (*precon)->options[LIS_OPTIONS_M] + 1; 

	for(i=0;i<n;i++)
	{
#ifdef _MPI
		lis_sort_iid(A->ptr[i],A->ptr[i+1]-1,A->index,A->g_index,A->value);
#else
		lis_sort_id(A->ptr[i],A->ptr[i+1]-1,A->index,A->value);
#endif
	}

	ptr  = (int *)lis_malloc((n+1)*sizeof(int));
	lis_matrix_create(n,0,A->comm,&C);

	for(i=0; i<n; i++)
	{
		for(j=A->ptr[i];j<A->ptr[i+1];j++)
		{
			if( i==A->index[j] )
			{
				diag[i] = j;
				break;
			}
		}
	}

	iw = (int *)lis_malloc( 2*gn*sizeof(int) );
	memset( iw,0,gn*sizeof(int) );
	iw2   = iw + gn;
	nnz   = 0;
	for(i=0;i<n;i++)
	{
		k = 0;
		n2 = _min(diag[i]+m,A->ptr[i+1]);
		for(ja=diag[i];ja<n2;ja++)
		{
			jj = A->index[ja];
			if( jj>=n ) continue;
			for(jb=B->ptr[jj];jb<B->ptr[jj+1];jb++)
			{
				jcol = B->index[jb];
				jpos = iw[jcol];
				if( jpos==0 )
				{
					iw2[k] = jcol;
					k++;
					nnz++;
					iw[jcol] = k;
				}
			}
		}
		ptr[i+1] = k;
		for(j=0;j<k;j++)
		{
			iw[iw2[j]] = 0;
		}
	}

	ptr[0] = 0;
	for(i=0;i<n;i++)
	{
		ptr[i+1] += ptr[i]; 
	}

	index  = (int *)lis_malloc(nnz*sizeof(int));
	value  = (LIS_SCALAR *)lis_malloc(nnz*sizeof(LIS_SCALAR));

	memset( iw,0,gn*sizeof(int) );
	for(i=0;i<n;i++)
	{
		k = ptr[i];
		t = 0.0;
		for(ja=diag[i];ja<_min(diag[i]+m,A->ptr[i+1]);ja++)
		{
			jj = A->index[ja];
			if( jj>=n ) continue;
			val = A->value[ja];
			if( jj!=i ) val = -w * val;
			for(jb=B->ptr[jj];jb<B->ptr[jj+1];jb++)
			{
				jcol = B->index[jb];
				jpos = iw[jcol];
				if( jpos==0 )
				{
					index[k]   = jcol;
					value[k]   = val * B->value[jb];
					k++;
					iw[jcol] = k;
				}
				else
				{
					value[jpos-1] += val * B->value[jb];
				}
			}
			t += val * b[jj];
		}
		for(j=ptr[i];j<k;j++)
		{
			iw[index[j]] = 0;
		}
		Pb[i] = t;
	}
	lis_matrix_set_crs(nnz,ptr,index,value,C);
	lis_matrix_assemble(&C,LIS_MATRIX_CRS);

	lis_matrix_scaling(C,Pbb,dd,LIS_SCALE_JACOBI);

	(*precon)->Pmat = C;
	(*precon)->Pb   = Pbb;
	lis_free(iw);
	lis_vector_destroy(dd);
#endif
	return LIS_SUCCESS;
}

void lis_psolve_is(LIS_PRECON precon, LIS_VECTOR b, LIS_VECTOR x)
{
	lis_vector_copy(b, x);
}

void lis_psolvet_is(LIS_PRECON precon, LIS_VECTOR b, LIS_VECTOR x)
{
	lis_vector_copy(b, x);
}


int lis_precon_create_is2(LIS_PRECON precon)
{
#ifdef _OPENMP
	int i,j,k,ik,m,id,threads;
	int n,nnz,ja,jj,jb,jcol,jpos,err;
	int *diag;
	int is,ie,n1,n2;
	LIS_SCALAR val,t;
	LIS_SCALAR w;
	LIS_SCALAR *d;
	LIS_MATRIX A;
	LIS_MATRIX_PARAMS paramm;
	double ctimes;

	A    = precon->Amat;
	diag = precon->diag;
	d    = precon->D->value;
	n    = A->n;
	w    = precon->params[LIS_PARAMS_ALPHA]; 
	m    = precon->options[LIS_OPTIONS_M] + 1; 


#pragma omp parallel for private(i,j)
	for(i=0; i<n; i++)
	{
		lis_sort_id(A->ptr[i],A->ptr[i+1]-1,A->index,A->value);
		for(j=A->ptr[i];j<A->ptr[i+1];j++)
		{
			if( i==A->index[j] )
			{
				diag[i] = j;
				d[i]    = A->value[j];
				break;
			}
		}
	}

#pragma omp parallel private(i,ja,jj,jb,jcol,val)
{
#pragma omp for
	for(i=0;i<n;i++)
	{
		for(ja=diag[i]+1;ja<_min(diag[i]+m,A->ptr[i+1]);ja++)
		{
			jj = A->index[ja];
			val = -w * A->value[ja];
			for(jb=A->ptr[jj];jb<A->ptr[jj+1];jb++)
			{
				jcol = A->index[jb];
				if( jcol==i )
				{
					d[i] += val * A->value[jb];
					break;
				}
			}
		}
		d[i] = 1.0 / d[i];
	}
}
#else
	int i,j,m;
	int n;
	int *diag;
	LIS_SCALAR w;
	LIS_SCALAR *d;
	LIS_MATRIX A;

	A    = precon->Amat;
	diag = precon->diag;
	d    = precon->D->value;
	n    = A->n;
	w    = precon->params[LIS_PARAMS_ALPHA]; 
	m    = precon->options[LIS_OPTIONS_M] + 1; 

	for(i=0;i<n;i++)
	{
#ifdef _MPI
		lis_sort_iid(A->ptr[i],A->ptr[i+1]-1,A->index,A->g_index,A->value);
#else
		lis_sort_id(A->ptr[i],A->ptr[i+1]-1,A->index,A->value);
#endif
	}

	for(i=0; i<n; i++)
	{
		for(j=A->ptr[i];j<A->ptr[i+1];j++)
		{
			if( i==A->index[j] )
			{
				diag[i] = j;
				d[i]    = A->value[j];
				break;
			}
		}
	}
	/*
	for(i=0;i<n;i++)
	{
		for(ja=diag[i]+1;ja<_min(diag[i]+m,A->ptr[i+1]);ja++)
		{
			jj = A->index[ja];
			val = -w * A->value[ja];
			for(jb=A->ptr[jj];jb<A->ptr[jj+1];jb++)
			{
				jcol = A->index[jb];
				if( jcol==i )
				{
					d[i] += val * A->value[jb];
					break;
				}
			}
		}
		d[i] = 1.0 / d[i];
	}
	*/
#endif
	return LIS_SUCCESS;
}


void lis_psolve_is_1(LIS_PRECON precon, LIS_VECTOR B, LIS_VECTOR X)
{
	int i,j,jj,n,m;
	int *diag;
	LIS_SCALAR t;
	LIS_SCALAR w;
	LIS_SCALAR *d;
	LIS_SCALAR *b,*x;
	LIS_MATRIX A;

	/*
	 *  x = M^{-1}b
	 *  M^{-1}  = (I+S)
	 */

	A = precon->Amat;
	d = precon->D->value;
	diag = precon->diag;
	b = B->value;
	x = X->value;
	n = A->n;
	w    = precon->params[LIS_PARAMS_ALPHA]; 
	m    = precon->options[LIS_OPTIONS_M] + 1; 

#ifdef _MPI
	if( A->np>B->np )
	{
		B->value = (LIS_SCALAR *)realloc(B->value,A->np*sizeof(LIS_SCALAR));
		if( B->value==NULL )
		{
			SETERR("out of memory\n");
			return;
		}
		b = B->value;
	}
	lis_send_recv(A->commtable,b);
#endif
#pragma omp parallel private(i,jj,j,t)
{
#pragma omp for
	for(i=0;i<n;i++)
	{
		t = 0.0;
		for(j=diag[i]+1;j<_min(diag[i]+m,A->ptr[i+1]);j++)
		{
			jj = A->index[j];
			t += A->value[j] * b[jj];
		}
		x[i] = b[i] - w*t;
//		x[i] = x[i] * d[i];
	}
}
}

void lis_psolve_is_2(LIS_PRECON precon, LIS_VECTOR B, LIS_VECTOR X)
{
	int i,j,jj,n,m;
	int *diag;
	LIS_SCALAR t;
	LIS_SCALAR w;
	LIS_SCALAR *d;
	LIS_SCALAR *y;
	LIS_SCALAR *b,*x;
	LIS_MATRIX A;

	/*
	 *  x = M^{-1}b
	 *  M^{-1}  = (I+S')(I+S)
	 */

#ifdef _OPENMP
	A = precon->Amat;
	d = precon->D->value;
	diag = precon->diag;
	b = B->value;
	x = B->value;
	y = precon->temp->value;
	n = A->n;
	w    = precon->params[LIS_PARAMS_ALPHA]; 
	m    = precon->options[LIS_OPTIONS_M] + 1; 

#pragma omp parallel private(i,jj,j,t)
{
#pragma omp for
	for(i=0;i<n;i++)
	{
		t = 0.0;
		for(j=diag[i]+1;j<_min(diag[i]+m,A->ptr[i+1]);j++)
		{
			jj = A->index[j];
			t += A->value[j] * b[jj];
		}
		y[i] = b[i] - w*t;
	}
#pragma omp for
	for(i=0;i<n;i++)
	{
		t = 0.0;
		for(j=diag[i]-1;j>=_max(diag[i]-m,A->ptr[i]);j--)
		{
			jj = A->index[j];
			t += A->value[j] * y[jj];
		}
		x[i] = y[i] - w*t;
//		x[i] = x[i] * d[i];
	}
}
#else
	A = precon->Amat;
	d = precon->D->value;
	diag = precon->diag;
	b = B->value;
	x = B->value;
	y = precon->temp->value;
	n = A->n;
	w    = precon->params[LIS_PARAMS_ALPHA]; 
	m    = precon->options[LIS_OPTIONS_M] + 1; 

#ifdef _MPI
	if( A->np>B->np )
	{
		B->value = (LIS_SCALAR *)realloc(B->value,A->np*sizeof(LIS_SCALAR));
		if( B->value==NULL )
		{
			SETERR("out of memory\n");
			return;
		}
		b = B->value;
	}
	lis_send_recv(A->commtable,b);
#endif
	for(i=0;i<n;i++)
	{
		t = 0.0;
		for(j=diag[i]+1;j<_min(diag[i]+m,A->ptr[i+1]);j++)
		{
			jj = A->index[j];
			t += A->value[j] * b[jj];
		}
		y[i] = b[i] - w*t;
	}
#ifdef _MPI
	if( A->np>precon->temp->np )
	{
		precon->temp->value = (LIS_SCALAR *)realloc(precon->temp->value,A->np*sizeof(LIS_SCALAR));
		if( precon->temp->value==NULL )
		{
			SETERR("out of memory\n");
			return;
		}
		y = precon->temp->value;
	}
	lis_send_recv(A->commtable,y);
#endif
	for(i=0;i<n;i++)
	{
		t = 0.0;
		for(j=diag[i]-1;j>=_max(diag[i]-m,A->ptr[i]);j--)
		{
			jj = A->index[j];
			t += A->value[j] * y[jj];
		}
		x[i] = y[i] - w*t;
//		x[i] = x[i] * d[i];
	}
#endif
}

void lis_psolvet_is_1(LIS_PRECON precon, LIS_VECTOR B, LIS_VECTOR X)
{
	int i,j,jj,n,m;
	int *diag;
	LIS_SCALAR t;
	LIS_SCALAR w;
	LIS_SCALAR *d;
	LIS_SCALAR *b,*x;
	LIS_MATRIX A;

	/*
	 *  x = M^{-1}b
	 *  M^{-1}  = (I+S)
	 */

	A = precon->At;
	d = precon->D->value;
	diag = precon->diag;
	b = B->value;
	x = X->value;
	n = A->n;
	w    = precon->params[LIS_PARAMS_ALPHA]; 
	m    = precon->options[LIS_OPTIONS_M] + 1; 

#ifdef _MPI
	if( A->np>B->np )
	{
		B->value = (LIS_SCALAR *)realloc(B->value,A->np*sizeof(LIS_SCALAR));
		if( B->value==NULL )
		{
			SETERR("out of memory\n");
			return;
		}
		b = B->value;
	}
	lis_send_recv(A->commtable,b);
#endif
#pragma omp parallel private(i,jj,j,t)
{
#pragma omp for
	for(i=0;i<n;i++)
	{
		t = 0.0;
		for(j=diag[i]+1;j<_min(diag[i]+m,A->ptr[i+1]);j++)
		{
			jj = A->index[j];
			t += A->value[j] * b[jj];
		}
		x[i] = b[i] - w*t;
//		x[i] = x[i] * d[i];
	}
}
}

void lis_psolvet_is_2(LIS_PRECON precon, LIS_VECTOR B, LIS_VECTOR X)
{
	int i,j,jj,n,m;
	int *diag;
	LIS_SCALAR t;
	LIS_SCALAR w;
	LIS_SCALAR *d;
	LIS_SCALAR *y;
	LIS_SCALAR *b,*x;
	LIS_MATRIX A;

	/*
	 *  x = M^{-1}b
	 *  M^{-1}  = (I+S')(I+S)
	 */

	A = precon->At;
	d = precon->D->value;
	diag = precon->diag;
	b = B->value;
	x = X->value;
	y = precon->temp->value;
	n = A->n;
	w    = precon->params[LIS_PARAMS_ALPHA]; 
	m    = precon->options[LIS_OPTIONS_M] + 1; 

#ifdef _OPENMP
#pragma omp parallel private(i,jj,j,t)
{
#pragma omp for
	for(i=0;i<n;i++)
	{
		t = 0.0;
		for(j=diag[i]+1;j<_min(diag[i]+m,A->ptr[i+1]);j++)
		{
			jj = A->index[j];
			t += A->value[j] * b[jj];
		}
		y[i] = b[i] - w*t;
	}
#pragma omp for
	for(i=0;i<n;i++)
	{
		t = 0.0;
		for(j=diag[i]-1;j>=_max(diag[i]-m,A->ptr[i]);j--)
		{
			jj = A->index[j];
			t += A->value[j] * y[jj];
		}
		x[i] = y[i] - w*t;
//		x[i] = x[i] * d[i];
	}
}
#else
#ifdef _MPI
	if( A->np>B->np )
	{
		B->value = (LIS_SCALAR *)realloc(B->value,A->np*sizeof(LIS_SCALAR));
		if( B->value==NULL )
		{
			SETERR("out of memory\n");
			return;
		}
		b = B->value;
	}
	lis_send_recv(A->commtable,b);
#endif
	for(i=0;i<n;i++)
	{
		t = 0.0;
		for(j=diag[i]+1;j<_min(diag[i]+m,A->ptr[i+1]);j++)
		{
			jj = A->index[j];
			t += A->value[j] * b[jj];
		}
		y[i] = b[i] - w*t;
	}
#ifdef _MPI
	if( A->np>precon->temp->np )
	{
		precon->temp->value = (LIS_SCALAR *)realloc(precon->temp->value,A->np*sizeof(LIS_SCALAR));
		if( precon->temp->value==NULL )
		{
			SETERR("out of memory\n");
			return;
		}
		y = precon->temp->value;
	}
	lis_send_recv(A->commtable,y);
#endif
	for(i=0;i<n;i++)
	{
		t = 0.0;
		for(j=diag[i]-1;j>=_max(diag[i]-m,A->ptr[i]);j--)
		{
			jj = A->index[j];
			t += A->value[j] * y[jj];
		}
		x[i] = y[i] - w*t;
//		x[i] = x[i] * d[i];
	}
#endif
}

