#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#ifdef _OPENMP
	#include <omp.h>
#endif
#include "lis.h"
#include "lis_precon.h"

#ifdef _OPENMP
int lis_precon_create_ssor(LIS_PRECON precon)
{
	int	i,j,n;
	LIS_SCALAR *D;

	n  = precon->Amat->n;
	D  = precon->D->value;
#pragma omp parallel for private(i,j)
	for(i=0; i<n; i++)
	{
		for(j=precon->Amat->ptr[i];j<precon->Amat->ptr[i+1];j++)
		{
			if( i==precon->Amat->index[j] )
			{
				precon->diag[i] = j;
				D[i] = 1.0 / precon->Amat->value[j];
				break;
			}
		}
	}
	return LIS_SUCCESS;
}

void lis_psolve_ssor(LIS_PRECON precon, LIS_VECTOR B, LIS_VECTOR X)
{
	int i,j,n;
	int n1,n2,k,jj,is,ie,threads;
	LIS_SCALAR w,t;
	LIS_SCALAR *b,*x,*d;

	/*
	 *  Mx = b
	 *  M  = (D/w + L) * (I + w*D^-1 * U)
	 */

	n = precon->Amat->n;
	w = precon->params[LIS_PARAMS_SSOR_W];
	b = B->value;
	x = X->value;
	d = precon->D->value;
	threads = omp_get_max_threads();
#pragma omp parallel for private(i,j,jj,is,ie,n1,n2,k,t)
	for(k=0;k<threads;k++)
	{
		n1 = n/threads;
		n2 = n%threads;
		if( k<n2 )
		{
			ie = n1+1;
			is = ie*k;
		}
		else
		{
			ie = n1;
			is = ie*k + n2;
		}
		ie = ie+is;

		for(i=is; i<ie; i++)
		{
			t = b[i];
			for(j=precon->Amat->ptr[i];j<precon->diag[i];j++)
			{
				jj = precon->Amat->index[j];
				if( jj>=is && jj<ie ) t -= precon->Amat->value[j] * x[jj];
//				t -= precon->Amat->value[j] * x[precon->Amat->index[j]];
			}
			x[i] = t * w * d[i];
		}	
		for(i=ie-1; i>=is; i--)
		{
			t = 0.0;
			for(j=precon->diag[i]+1;j<precon->Amat->ptr[i+1];j++)
			{
				jj = precon->Amat->index[j];
				if( jj>=is && jj<ie ) t += precon->Amat->value[j] * x[jj];
//				t += precon->Amat->value[j] * x[precon->Amat->index[j]];
			}
			x[i] -= t * w * d[i];
		}
	}
}

void lis_psolvet_ssor(LIS_PRECON precon, LIS_VECTOR B, LIS_VECTOR X)
{
	int i,j,n;
	int n1,n2,k,jj,is,ie,threads;
	LIS_SCALAR w,t;
	LIS_SCALAR *b,*x,*d;

	/*
	 *  M'x = b
	 *  M'  = (I + U' * w*D^-1) * (D/w + L')
	 */

	n = precon->Amat->n;
	w = precon->params[LIS_PARAMS_SSOR_W];
	b = B->value;
	x = X->value;
	d = precon->D->value;
	threads = omp_get_max_threads();
#pragma omp parallel for private(i,j,jj,is,ie,n1,n2,k,t)
	for(k=0;k<threads;k++)
	{
		n1 = n/threads;
		n2 = n%threads;
		if( k<n2 )
		{
			ie = n1+1;
			is = ie*k;
		}
		else
		{
			ie = n1;
			is = ie*k + n2;
		}
		ie = ie+is;

		for(i=is; i<ie; i++)
		{
			x[i] = b[i];
		}
		for(i=is; i<ie; i++)
		{
			t = x[i] * w * d[i];
			for(j=precon->diag[i]+1;j<precon->Amat->ptr[i+1];j++)
			{
				jj = precon->Amat->index[j];
				if( jj>=is && jj<ie ) x[jj] -= precon->Amat->value[j] * t;
//				x[precon->Amat->index[j]] -= precon->Amat->value[j] * t;
			}
		}
		for(i=ie-1; i>=is; i--)
		{
			t    = x[i] * w * d[i];
			x[i] = t;
			for(j=precon->Amat->ptr[i];j<precon->diag[i];j++)
			{
				jj = precon->Amat->index[j];
				if( jj>=is && jj<ie ) x[jj] -= precon->Amat->value[j] * t;
//				x[precon->Amat->index[j]] -= precon->Amat->value[j] * t;
			}
		}
	}
}

#else

int lis_precon_create_ssor(LIS_PRECON precon)
{
	int	i,j,n;
	LIS_SCALAR *D;

	n  = precon->Amat->n;
	D  = precon->D->value;

	for(i=0; i<n; i++)
	{
		for(j=precon->Amat->ptr[i];j<precon->Amat->ptr[i+1];j++)
		{
			if( i==precon->Amat->index[j] )
			{
				precon->diag[i] = j;
				D[i] = 1.0 / precon->Amat->value[j];
				break;
			}
		}
	}
	return LIS_SUCCESS;
}

void lis_psolve_ssor(LIS_PRECON precon, LIS_VECTOR B, LIS_VECTOR X)
{
	int i,j,jj,n;
	LIS_SCALAR w,t;
	LIS_SCALAR *b,*x,*d;

	/*
	 *  Mx = b
	 *  M  = (D/w + L) * (I + w*D^-1 * U)
	 */

	n = precon->Amat->n;
	w = precon->params[LIS_PARAMS_SSOR_W];
	b = B->value;
	x = X->value;
	d = precon->D->value;
	for(i=0; i<n; i++)
	{
		t = b[i];
		for(j=precon->Amat->ptr[i];j<precon->diag[i];j++)
		{
			jj = precon->Amat->index[j];
			if( jj<n ) t -= precon->Amat->value[j] * x[jj];
//			t -= precon->Amat->value[j] * x[jj];
		}
		x[i] = t * w * d[i];
	}
	for(i=n-1; i>=0; i--)
	{
		t = 0.0;
		for(j=precon->diag[i]+1;j<precon->Amat->ptr[i+1];j++)
		{
			jj = precon->Amat->index[j];
			if( jj<n ) t += precon->Amat->value[j] * x[jj];
//			t += precon->Amat->value[j] * x[jj];
		}
		x[i] -= t * w * d[i];
	}
}

void lis_psolvet_ssor(LIS_PRECON precon, LIS_VECTOR B, LIS_VECTOR X)
{
	int i,j,jj,n;
	LIS_SCALAR w,t;
	LIS_SCALAR *b,*x,*d;

	/*
	 *  M'x = b
	 *  M'  = (I + U' * w*D^-1) * (D/w + L')
	 */

	n = precon->Amat->n;
	w = precon->params[LIS_PARAMS_SSOR_W];
	b = B->value;
	x = X->value;
	d = precon->D->value;
	for(i=0; i<n; i++)
	{
		x[i] = b[i];
	}
	for(i=0; i<n; i++)
	{
		t = x[i] * w * d[i];
		for(j=precon->diag[i]+1;j<precon->Amat->ptr[i+1];j++)
		{
			jj = precon->Amat->index[j];
			if( jj<n ) x[jj] -= precon->Amat->value[j] * t;
//			x[jj] -= precon->Amat->value[j] * t;
		}
	}
	for(i=n-1; i>=0; i--)
	{
		t    = x[i] * w * d[i];
		x[i] = t;
		for(j=precon->Amat->ptr[i];j<precon->diag[i];j++)
		{
			jj = precon->Amat->index[j];
			if( jj<n ) x[jj] -= precon->Amat->value[j] * t;
//			x[jj] -= precon->Amat->value[j] * t;
		}
	}
}
#endif
